#!/usr/bin/env python3

#    Copyright (C) 2013-2017 Christian Thomas Jacobs.

#    This file is part of PyQSO.

#    PyQSO is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyQSO is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyQSO.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, Gdk
import logging
import os
try:
    import configparser
except ImportError:
    import ConfigParser as configparser
from datetime import datetime
from os.path import expanduser
import base64
try:
    import Hamlib
    have_hamlib = True
except ImportError:
    logging.warning("Could not import the Hamlib module!")
    have_hamlib = False

from pyqso.adif import *
from pyqso.callsign_lookup import *
from pyqso.auxiliary_dialogs import *
from pyqso.calendar_dialog import CalendarDialog


class RecordDialog:

    """ A dialog through which users can enter information about a QSO/record. """

    def __init__(self, application, log, index=None):
        """ Set up the layout of the record dialog, populate the various fields with the QSO details (if the record already exists), and show the dialog to the user.

        :arg application: The PyQSO application containing the main Gtk window, etc.
        :arg log: The log to which the record belongs (or will belong).
        :arg int index: If specified, then the dialog turns into 'edit record mode' and fills the data sources (e.g. the Gtk.Entry boxes) with the existing data in the log. If not specified (i.e. index is None), then the dialog starts off with nothing in the data sources.
        """

        logging.debug("Setting up the record dialog...")

        self.application = application
        self.builder = self.application.builder
        glade_file_path = os.path.join(os.path.realpath(os.path.dirname(__file__)), "res", "pyqso.glade")
        self.builder.add_objects_from_file(glade_file_path, ("record_dialog",))
        self.dialog = self.builder.get_object("record_dialog")
        self.builder.get_object("record_dialog").connect("key-press-event", self.on_key_press)

        # Set dialog title
        if(index is not None):
            self.dialog.set_title("Edit Record %d" % index)
        else:
            self.dialog.set_title("Add Record")

        # Check if a configuration file is present, since we might need it to set up the rest of the dialog.
        config = configparser.ConfigParser()
        have_config = (config.read(expanduser('~/.config/pyqso/preferences.ini')) != [])

        # Create label:entry pairs and store them in a dictionary
        self.sources = {}

        # QSO INFORMATION

        # CALL
        self.sources["CALL"] = self.builder.get_object("qso_call_entry")
        self.builder.get_object("callsign_lookup").connect("clicked", self.callsign_lookup_callback)

        # DATE
        self.sources["QSO_DATE"] = self.builder.get_object("qso_date_entry")
        self.builder.get_object("select_date").connect("clicked", self.calendar_callback)

        # TIME
        self.sources["TIME_ON"] = self.builder.get_object("qso_time_entry")
        self.builder.get_object("current_datetime").connect("clicked", self.set_current_datetime_callback)

        # FREQ
        self.sources["FREQ"] = self.builder.get_object("qso_frequency_entry")
        (section, option) = ("records", "default_frequency_unit")
        if(have_config and config.has_option(section, option)):
            self.frequency_unit = config.get(section, option)
            self.builder.get_object("qso_frequency_label").set_label("Frequency (%s)" % self.frequency_unit)
        else:
            self.frequency_unit = "MHz"

        # BAND
        self.sources["BAND"] = self.builder.get_object("qso_band_combo")
        for band in BANDS:
            self.sources["BAND"].append_text(band)
        self.sources["BAND"].set_active(0)  # Set an empty string as the default option.

        # MODE
        self.sources["MODE"] = self.builder.get_object("qso_mode_combo")
        for mode in sorted(MODES.keys()):
            self.sources["MODE"].append_text(mode)
        self.sources["MODE"].set_active(0)  # Set an empty string as the default option.
        self.sources["MODE"].connect("changed", self.on_mode_changed)

        # SUBMODE
        self.sources["SUBMODE"] = self.builder.get_object("qso_submode_combo")
        self.sources["SUBMODE"].append_text("")
        self.sources["SUBMODE"].set_active(0)  # Set an empty string initially. As soon as the user selects a particular MODE, the available SUBMODES will appear.

        # PROP_MODE
        self.sources["PROP_MODE"] = self.builder.get_object("qso_propagation_mode_combo")
        for propagation_mode in PROPAGATION_MODES:
            self.sources["PROP_MODE"].append_text(propagation_mode)
        self.sources["PROP_MODE"].set_active(0)  # Set an empty string as the default option.

        # POWER
        self.sources["TX_PWR"] = self.builder.get_object("qso_power_entry")

        # RST_SENT
        self.sources["RST_SENT"] = self.builder.get_object("qso_rst_sent_entry")

        # RST_RCVD
        self.sources["RST_RCVD"] = self.builder.get_object("qso_rst_received_entry")

        # QSL_SENT
        self.sources["QSL_SENT"] = self.builder.get_object("qso_qsl_sent_combo")
        qsl_sent_options = ["", "Y", "N", "R", "Q", "I"]
        for option in qsl_sent_options:
            self.sources["QSL_SENT"].append_text(option)
        self.sources["QSL_SENT"].set_active(0)  # Set an empty string as the default option.

        # QSL_RCVD
        self.sources["QSL_RCVD"] = self.builder.get_object("qso_qsl_received_combo")
        qsl_rcvd_options = ["", "Y", "N", "R", "I", "V"]
        for option in qsl_rcvd_options:
            self.sources["QSL_RCVD"].append_text(option)
        self.sources["QSL_RCVD"].set_active(0)  # Set an empty string as the default option.

        # NOTES
        self.sources["NOTES"] = self.builder.get_object("qso_notes_textview").get_buffer()

        # STATION INFORMATION

        # NAME
        self.sources["NAME"] = self.builder.get_object("station_name_entry")

        # ADDRESS
        self.sources["ADDRESS"] = self.builder.get_object("station_address_entry")

        # STATE
        self.sources["STATE"] = self.builder.get_object("station_state_entry")

        # COUNTRY
        self.sources["COUNTRY"] = self.builder.get_object("station_country_entry")

        # DXCC
        self.sources["DXCC"] = self.builder.get_object("station_dxcc_entry")

        # CQZ
        self.sources["CQZ"] = self.builder.get_object("station_cq_entry")

        # ITUZ
        self.sources["ITUZ"] = self.builder.get_object("station_itu_entry")

        # IOTA
        self.sources["IOTA"] = self.builder.get_object("station_iota_entry")

        # GRIDSQUARE
        self.sources["GRIDSQUARE"] = self.builder.get_object("station_gridsquare_entry")

        # SATELLITE INFORMATION

        # SAT_NAME
        self.sources["SAT_NAME"] = self.builder.get_object("satellite_name_entry")

        # SAT_MODE
        self.sources["SAT_MODE"] = self.builder.get_object("satellite_mode_entry")

        # Populate various fields, if possible.
        if(index is not None):
            # The record already exists, so display its current data in the input boxes.
            record = log.get_record_by_index(index)
            field_names = AVAILABLE_FIELD_NAMES_ORDERED
            for i in range(0, len(field_names)):
                data = record[field_names[i].lower()]
                if(data is None):
                    data = ""
                if(field_names[i] == "BAND"):
                    self.sources[field_names[i]].set_active(BANDS.index(data))
                elif(field_names[i] == "FREQ" and self.frequency_unit != "MHz"):
                    converted = self.convert_frequency(data, from_unit="MHz", to_unit=self.frequency_unit)
                    self.sources[field_names[i]].set_text(str(converted))
                elif(field_names[i] == "MODE"):
                    self.sources[field_names[i]].set_active(sorted(MODES.keys()).index(data))
                    # Handle SUBMODE at the same time.
                    submode_data = record["submode"]
                    if(submode_data is None):
                        submode_data = ""
                    self.sources["SUBMODE"].set_active(MODES[data].index(submode_data))
                elif(field_names[i] == "SUBMODE"):
                    # Skip, because this has been (or will be) handled when populating the MODE field.
                    continue
                elif(field_names[i] == "PROP_MODE"):
                    self.sources[field_names[i]].set_active(PROPAGATION_MODES.index(data))
                elif(field_names[i] == "QSL_SENT"):
                    self.sources[field_names[i]].set_active(qsl_sent_options.index(data))
                elif(field_names[i] == "QSL_RCVD"):
                    self.sources[field_names[i]].set_active(qsl_rcvd_options.index(data))
                else:
                    self.sources[field_names[i]].set_text(data)
        else:
            # Automatically fill in the current date and time
            self.set_current_datetime_callback()

            # Set up default field values
            # Mode
            (section, option) = ("records", "default_mode")
            if(have_config and config.has_option(section, option)):
                mode = config.get(section, option)
            else:
                mode = ""
            self.sources["MODE"].set_active(sorted(MODES.keys()).index(mode))

            # Submode
            (section, option) = ("records", "default_submode")
            if(have_config and config.has_option(section, option)):
                submode = config.get(section, option)
            else:
                submode = ""
            self.sources["SUBMODE"].set_active(MODES[mode].index(submode))

            # Power
            (section, option) = ("records", "default_power")
            if(have_config and config.has_option(section, option)):
                power = config.get(section, option)
            else:
                power = ""
            self.sources["TX_PWR"].set_text(power)

            # If the Hamlib module is present, then use it to fill in various fields if desired.
            if(have_hamlib):
                if(have_config and config.has_option("hamlib", "autofill") and config.has_option("hamlib", "rig_model") and config.has_option("hamlib", "rig_pathname")):
                    autofill = (config.getboolean("hamlib", "autofill"))
                    rig_model = config.get("hamlib", "rig_model")
                    rig_pathname = config.get("hamlib", "rig_pathname")
                    if(autofill):
                        self.hamlib_autofill(rig_model, rig_pathname)

        # Do we want PyQSO to autocomplete the Band field based on the Frequency field?
        (section, option) = ("records", "autocomplete_band")
        if(have_config and config.has_option(section, option)):
            autocomplete_band = (config.getboolean(section, option))
            if(autocomplete_band):
                self.sources["FREQ"].connect("changed", self.autocomplete_band)
        else:
            # If no configuration file exists, autocomplete the Band field by default.
            self.sources["FREQ"].connect("changed", self.autocomplete_band)

        self.dialog.show_all()

        logging.debug("Record dialog ready!")

        return

    def get_data(self, field_name):
        """ Return the data for a specified field from the Gtk.Entry/Gtk.ComboBoxText/etc boxes in the record dialog.

        :arg str field_name: The name of the field containing the desired data.
        :returns: The data in the specified field.
        :rtype: str
        """
        logging.debug("Retrieving the data in field %s from the record dialog..." % field_name)
        if(field_name == "CALL"):
            # Always show the callsigns in upper case.
            return self.sources[field_name].get_text().upper()
        elif(field_name == "FREQ" and self.frequency_unit != "MHz"):
            converted = self.convert_frequency(self.sources[field_name].get_text(), from_unit=self.frequency_unit, to_unit="MHz")
            return str(converted)
        elif(field_name == "MODE"):
            return self.sources["MODE"].get_active_text()
        elif(field_name == "SUBMODE"):
            return self.sources["SUBMODE"].get_active_text()
        elif(field_name == "PROP_MODE"):
            return self.sources["PROP_MODE"].get_active_text()
        elif(field_name == "BAND" or field_name == "QSL_SENT" or field_name == "QSL_RCVD"):
            return self.sources[field_name].get_active_text()
        elif(field_name == "NOTES"):
            (start, end) = self.sources[field_name].get_bounds()
            text = self.sources[field_name].get_text(start, end, True)
            return text
        else:
            return self.sources[field_name].get_text()

    def on_mode_changed(self, combo):
        """ If the MODE field has changed its value, then fill the SUBMODE field with all the available SUBMODE options for that new MODE. """
        self.sources["SUBMODE"].get_model().clear()
        mode = combo.get_active_text()
        for submode in MODES[mode]:
            self.sources["SUBMODE"].append_text(submode)
        self.sources["SUBMODE"].set_active(MODES[mode].index(""))  # Set the submode to an empty string.
        return

    def on_key_press(self, widget, event):
        """ If the Return key is pressed, emit the "OK" response to record the QSO. """
        child = widget.get_focus()
        if(not(isinstance(child, Gtk.ToggleButton) or isinstance(child, Gtk.Button) or isinstance(child, Gtk.TextView)) and event.keyval == Gdk.KEY_Return):
            self.dialog.emit('response', Gtk.ResponseType.OK)
        return

    def autocomplete_band(self, widget=None):
        """ If a value for the Frequency is entered, this function autocompletes the Band field. """

        frequency = self.sources["FREQ"].get_text()

        # Check whether we actually have a (valid) value to use. If not, set the BAND field to an empty string ("").
        try:
            frequency = float(frequency)
        except ValueError:
            self.sources["BAND"].set_active(0)
            return

        # Convert to MHz if necessary.
        if(self.frequency_unit != "MHz"):
            frequency = self.convert_frequency(frequency, from_unit=self.frequency_unit, to_unit="MHz")

        # Find which band the frequency lies in.
        for i in range(1, len(BANDS)):
            if(frequency >= BANDS_RANGES[i][0] and frequency <= BANDS_RANGES[i][1]):
                self.sources["BAND"].set_active(i)
                return

        self.sources["BAND"].set_active(0)  # If we've reached this, then the frequency does not lie in any of the specified bands.
        return

    def hamlib_autofill(self, rig_model, rig_pathname):
        """ Set the various fields using data from the radio via Hamlib.

        :arg str rig_model: The model of the radio/rig.
        :arg str rig_pathname: The path to the rig (or rig control device).
        """

        # Open a communication channel to the radio.
        try:
            Hamlib.rig_set_debug(Hamlib.RIG_DEBUG_NONE)
            rig = Hamlib.Rig(Hamlib.__dict__[rig_model])  # Look up the model's numerical index in Hamlib's symbol dictionary.
            rig.set_conf("rig_pathname", rig_pathname)
            rig.open()
        except:
            logging.error("Could not open a communication channel to the rig via Hamlib!")
            return

        # Frequency
        try:
            frequency = "%.6f" % (rig.get_freq()/1.0e6)  # Converting to MHz here.
            # Convert to the desired unit, if necessary.
            if(self.frequency_unit != "MHz"):
                frequency = str(self.convert_frequency(frequency, from_unit="MHz", to_unit=self.frequency_unit))
            self.sources["FREQ"].set_text(frequency)
        except:
            logging.error("Could not obtain the current frequency via Hamlib!")

        # Mode
        try:
            (mode, width) = rig.get_mode()
            mode = Hamlib.rig_strrmode(mode).upper()
            # Handle USB and LSB as special cases.
            if(mode == "USB" or mode == "LSB"):
                submode = mode
                mode = "SSB"
                self.sources["MODE"].set_active(sorted(MODES.keys()).index(mode))
                self.sources["SUBMODE"].set_active(MODES[mode].index(submode))
            else:
                self.sources["MODE"].set_active(sorted(MODES.keys()).index(mode))
        except:
            logging.error("Could not obtain the current mode (e.g. FM, AM, CW) via Hamlib!")

        # Close communication channel.
        try:
            rig.close()
        except:
            logging.error("Could not close the communication channel to the rig via Hamlib!")

        return

    def callsign_lookup_callback(self, widget=None):
        """ Get the callsign-related data from an online database and store it in the relevant Gtk.Entry boxes, but return None. """

        # Get the database name.
        config = configparser.ConfigParser()
        have_config = (config.read(expanduser('~/.config/pyqso/preferences.ini')) != [])
        try:
            if(have_config and config.has_option("records", "callsign_database")):
                database = config.get("records", "callsign_database")
                if(database == ""):
                    raise ValueError
            else:
                raise ValueError
        except ValueError:
            error(parent=self.dialog, message="To perform a callsign lookup, please specify the name of the callsign database in the Preferences.")
            return

        try:
            if(database == "qrz.com"):
                # QRZ.com
                callsign_lookup = CallsignLookupQRZ(parent=self.dialog)
            elif(database == "hamqth.com"):
                # HamQTH.com
                callsign_lookup = CallsignLookupHamQTH(parent=self.dialog)
            else:
                raise ValueError("Unknown callsign database: %s" % database)
        except ValueError as e:
            logging.exception(e)
            error(parent=self.dialog, message=e)
            return

        # Get username and password from configuration file.
        if(have_config and config.has_option("records", "callsign_database_username") and config.has_option("records", "callsign_database_password")):
            username = config.get("records", "callsign_database_username")
            password = base64.b64decode(config.get("records", "callsign_database_password")).decode("utf-8")
            if(not username or not password):
                details_given = False
            else:
                details_given = True
        else:
            details_given = False
        if(not details_given):
            error(parent=self.dialog, message="To perform a callsign lookup, please specify your username and password in the Preferences.")
            return

        # Get the callsign from the CALL field.
        full_callsign = self.sources["CALL"].get_text()
        if(not full_callsign):
            # Empty callsign field.
            error(parent=self.dialog, message="Please enter a callsign to lookup.")
            return

        # Connect to the database.
        connected = callsign_lookup.connect(username, password)
        if(connected):
            # Check whether we want to ignore any prefixes (e.g. "IA/") or suffixes "(e.g. "/M") in the callsign
            # before performing the lookup.
            if(have_config and config.has_option("records", "ignore_prefix_suffix")):
                ignore_prefix_suffix = (config.getboolean("records", "ignore_prefix_suffix"))
            else:
                ignore_prefix_suffix = True

            # Perform the lookup.
            fields_and_data = callsign_lookup.lookup(full_callsign, ignore_prefix_suffix=ignore_prefix_suffix)
            for field_name in list(fields_and_data.keys()):
                self.sources[field_name].set_text(fields_and_data[field_name])
        return

    def calendar_callback(self, widget):
        """ Open up a calendar widget for easy QSO_DATE selection. Return None after the user destroys the dialog. """
        c = CalendarDialog(self.application)
        response = c.dialog.run()
        if(response == Gtk.ResponseType.OK):
            self.sources["QSO_DATE"].set_text(c.date)
        c.dialog.destroy()
        return

    def set_current_datetime_callback(self, widget=None):
        """ Insert the current date and time. """

        # Check if a configuration file is present.
        config = configparser.ConfigParser()
        have_config = (config.read(expanduser('~/.config/pyqso/preferences.ini')) != [])

        # Do we want to use UTC or the computer's local time?
        (section, option) = ("records", "use_utc")
        if(have_config and config.has_option(section, option)):
            use_utc = (config.getboolean(section, option))
            if(use_utc):
                dt = datetime.utcnow()
            else:
                dt = datetime.now()
        else:
            dt = datetime.utcnow()  # Use UTC by default, since this is expected by ADIF.

        self.sources["QSO_DATE"].set_text(dt.strftime("%Y%m%d"))
        self.sources["TIME_ON"].set_text(dt.strftime("%H%M"))

        return

    def convert_frequency(self, frequency, from_unit, to_unit):
        """ Convert a frequency from one unit to another.

        :arg float frequency: The frequency to convert.
        :arg str from_unit: The current unit of the frequency.
        :arg str to_unit: The desired unit of the frequency.
        :rtype: float
        :returns: The frequency in the to_unit.
        """
        scaling = {"Hz": 1, "kHz": 1e3, "MHz": 1e6, "GHz": 1e9}
        # Check that the from/to frequency units are valid.
        try:
            if(from_unit not in scaling.keys()):
                raise ValueError("Unknown frequency unit '%s' in from_unit" % from_unit)
            if(to_unit not in scaling.keys()):
                raise ValueError("Unknown frequency unit '%s' in to_unit" % to_unit)
        except ValueError as e:
            logging.exception(e)
            return frequency
        # Cast to float before scaling.
        if(not isinstance(frequency, float)):
            try:
                if(frequency == "" or frequency is None):
                    return frequency
                else:
                    frequency = float(frequency)
            except(ValueError, TypeError):
                logging.exception("Could not convert frequency to a floating-point value.")
                return frequency
        # Do not bother scaling if the units are the same.
        if(from_unit == to_unit):
            return frequency

        coefficient = scaling[from_unit]/scaling[to_unit]
        return float("%.6f" % (coefficient*frequency))



#!/usr/bin/env python3

#    Copyright (C) 2017 Christian Thomas Jacobs.

#    This file is part of PyQSO.

#    PyQSO is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyQSO is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyQSO.  If not, see <http://www.gnu.org/licenses/>.


def compare_date_and_time(model, row1, row2, user_data):
    """ Compare two rows (let's call them A and B) in a Gtk.ListStore, and sort by both date and time.

    :arg Gtk.TreeModel model: The model used to sort the log data.
    :arg Gtk.TreeIter row1: The pointer to row A.
    :arg Gtk.TreeIter row2: The pointer to row B.
    :arg user_data: The specific column from which to retrieve data for rows A and B.
    :returns: -1 if Row B's date/time is more recent than Row A's; 0 if both dates and times are the same; 1 if Row A's date/time is more recent than Row B's.
    :rtype: int
    """
    date1 = model.get_value(row1, user_data[0])
    date2 = model.get_value(row2, user_data[0])
    time1 = model.get_value(row1, user_data[1])
    time2 = model.get_value(row2, user_data[1])
    if(date1 < date2):
        return -1
    elif(date1 == date2):
        # If the dates are the same, then let's also sort by time.
        if(time1 > time2):
            return 1
        elif(time1 == time2):
            return 0
        else:
            return -1
    else:
        return 1


def compare_default(model, row1, row2, user_data):
    """ The default sorting function for all Gtk.ListStore objects.

    :arg Gtk.TreeModel model: The model used to sort the log data.
    :arg Gtk.TreeIter row1: The pointer to row A.
    :arg Gtk.TreeIter row2: The pointer to row B.
    :arg user_data: The specific column from which to retrieve data for rows A and B.
    :returns: -1 if the value of Row A's column value is less than Row B's column value; 0 if both values are the same; 1 if Row A's column value is greater than Row B's column value.
    :rtype: int
    """

    # Let's try to deal with numerical values, if possible.
    try:
        value1 = float(model.get_value(row1, user_data))
        value2 = float(model.get_value(row2, user_data))
    except ValueError:
        value1 = model.get_value(row1, user_data)
        value2 = model.get_value(row2, user_data)

    if(value1 < value2):
        return -1
    elif(value1 == value2):
        return 0
    else:
        return 1



#!/usr/bin/env python3

#    Copyright (C) 2012-2017 Christian Thomas Jacobs.

#    This file is part of PyQSO.

#    PyQSO is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyQSO is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyQSO.  If not, see <http://www.gnu.org/licenses/>.

import re
import logging
from datetime import datetime
import calendar
try:
    import configparser
except ImportError:
    import ConfigParser as configparser
from os.path import expanduser

# ADIF field names and their associated data types available in PyQSO.
AVAILABLE_FIELD_NAMES_TYPES = {"CALL": "S",
                               "QSO_DATE": "D",
                               "TIME_ON": "T",
                               "FREQ": "N",
                               "BAND": "E",
                               "MODE": "E",
                               "SUBMODE": "E",
                               "PROP_MODE": "E",
                               "TX_PWR": "N",
                               "RST_SENT": "S",
                               "RST_RCVD": "S",
                               "QSL_SENT": "S",
                               "QSL_RCVD": "S",
                               "NOTES": "M",
                               "NAME": "S",
                               "ADDRESS": "S",
                               "STATE": "S",
                               "COUNTRY": "S",
                               "DXCC": "N",
                               "CQZ": "N",
                               "ITUZ": "N",
                               "IOTA": "C",
                               "GRIDSQUARE": "S",
                               "SAT_NAME": "S",
                               "SAT_MODE": "S"}
# Note: The logbook uses the ADIF field names for the database column names.
# This list is used to display the columns in a logical order.
AVAILABLE_FIELD_NAMES_ORDERED = ["CALL", "QSO_DATE", "TIME_ON", "FREQ", "BAND", "MODE", "SUBMODE", "PROP_MODE", "TX_PWR",
                                 "RST_SENT", "RST_RCVD", "QSL_SENT", "QSL_RCVD", "NOTES", "NAME",
                                 "ADDRESS", "STATE", "COUNTRY", "DXCC", "CQZ", "ITUZ", "IOTA", "GRIDSQUARE", "SAT_NAME", "SAT_MODE"]
# Define the more user-friendly versions of the field names.
AVAILABLE_FIELD_NAMES_FRIENDLY = {"CALL": "Callsign",
                                  "QSO_DATE": "Date",
                                  "TIME_ON": "Time",
                                  "FREQ": "Frequency (MHz)",
                                  "BAND": "Band",
                                  "MODE": "Mode",
                                  "SUBMODE": "Submode",
                                  "PROP_MODE": "Propagation Mode",
                                  "TX_PWR": "TX Power (W)",
                                  "RST_SENT": "RST Sent",
                                  "RST_RCVD": "RST Received",
                                  "QSL_SENT": "QSL Sent",
                                  "QSL_RCVD": "QSL Received",
                                  "NOTES": "Notes",
                                  "NAME": "Name",
                                  "ADDRESS": "Address",
                                  "STATE": "State",
                                  "COUNTRY": "Country",
                                  "DXCC": "DXCC",
                                  "CQZ": "CQ Zone",
                                  "ITUZ": "ITU Zone",
                                  "IOTA": "IOTA Designator",
                                  "GRIDSQUARE": "Grid Square",
                                  "SAT_NAME": "Satellite Name",
                                  "SAT_MODE": "Satellite Mode"}

# A: AwardList
# B: Boolean
# N: Number
# S: String
# I: International string
# D: Date
# T: Time
# M: Multi-line string
# G: Multi-line international string
# L: Location
# E: Enumerated
DATA_TYPES = ["A", "B", "N", "S", "I", "D", "T", "M", "G", "L", "E"]

# All the valid modes listed in the ADIF specification. This is a dictionary with the key-value pairs holding the MODE and SUBMODE(s) respectively.
MODES = {"": ("",),
         "AM": ("",),
         "ATV": ("",),
         "CHIP": ("", "CHIP64", "CHIP128"),
         "CLO": ("",),
         "CONTESTI": ("",),
         "CW": ("", "PCW"),
         "DIGITALVOICE": ("",),
         "DOMINO": ("", "DOMINOEX", "DOMINOF"),
         "DSTAR": ("",),
         "FAX": ("",),
         "FM": ("",),
         "FSK441": ("",),
         "FT8": ("",),
         "HELL": ("", "FMHELL", "FSKHELL", "HELL80", "HFSK", "PSKHELL"),
         "ISCAT": ("", "ISCAT-A", "ISCAT-B"),
         "JT4": ("", "JT4A", "JT4B", "JT4C", "JT4D", "JT4E", "JT4F", "JT4G"),
         "JT6M": ("",),
         "JT9": ("",),
         "JT44": ("",),
         "JT65": ("", "JT65A", "JT65B", "JT65B2", "JT65C", "JT65C2"),
         "MFSK": ("", "MFSK4", "MFSK8", "MFSK11", "MFSK16", "MFSK22", "MFSK31", "MFSK32", "MFSK64", "MFSK128"),
         "MT63": ("",),
         "OLIVIA": ("", "OLIVIA 4/125", "OLIVIA 4/250", "OLIVIA 8/250", "OLIVIA 8/500", "OLIVIA 16/500", "OLIVIA 16/1000", "OLIVIA 32/1000"),
         "OPERA": ("", "OPERA-BEACON", "OPERA-QSO"),
         "PAC": ("", "PAC2", "PAC3", "PAC4"),
         "PAX": ("", "PAX2"),
         "PKT": ("",),
         "PSK": ("", "FSK31", "PSK10", "PSK31", "PSK63", "PSK63F", "PSK125", "PSK250", "PSK500", "PSK1000", "PSKAM10", "PSKAM31", "PSKAM50", "PSKFEC31", "QPSK31", "QPSK63", "QPSK125", "QPSK250", "QPSK500"),
         "PSK2K": ("",),
         "Q15": ("",),
         "ROS": ("", "ROS-EME", "ROS-HF", "ROS-MF"),
         "RTTY": ("", "ASCI"),
         "RTTYM": ("",),
         "SSB": ("", "LSB", "USB"),
         "SSTV": ("",),
         "THOR": ("",),
         "THRB": ("", "THRBX"),
         "TOR": ("", "AMTORFEC", "GTOR"),
         "V4": ("",),
         "VOI": ("",),
         "WINMOR": ("",),
         "WSPR": ("",)
         }

# A dictionary of all the deprecated MODE values.
MODES_DEPRECATED = {"AMTORFEC": ("",),
                    "ASCI": ("",),
                    "CHIP64": ("",),
                    "CHIP128": ("",),
                    "DOMINOF": ("",),
                    "FMHELL": ("",),
                    "FSK31": ("",),
                    "GTOR": ("",),
                    "HELL80": ("",),
                    "HFSK": ("",),
                    "JT4A": ("",),
                    "JT4B": ("",),
                    "JT4C": ("",),
                    "JT4D": ("",),
                    "JT4E": ("",),
                    "JT4F": ("",),
                    "JT4G": ("",),
                    "JT65A": ("",),
                    "JT65B": ("",),
                    "JT65C": ("",),
                    "MFSK8": ("",),
                    "MFSK16": ("",),
                    "PAC2": ("",),
                    "PAC3": ("",),
                    "PAX2": ("",),
                    "PCW": ("",),
                    "PSK10": ("",),
                    "PSK31": ("",),
                    "PSK63": ("",),
                    "PSK63F": ("",),
                    "PSK125": ("",),
                    "PSKAM10": ("",),
                    "PSKAM31": ("",),
                    "PSKAM50": ("",),
                    "PSKFEC31": ("",),
                    "PSKHELL": ("",),
                    "QPSK31": ("",),
                    "QPSK63": ("",),
                    "QPSK125": ("",),
                    "THRBX": ("",)
                    }

# Include all deprecated modes.
MODES.update(MODES_DEPRECATED)

# All the bands listed in the ADIF specification.
BANDS = ["", "2190m", "630m", "560m", "160m", "80m", "60m", "40m", "30m", "20m", "17m", "15m", "12m", "10m", "6m", "4m", "2m", "1.25m", "70cm", "33cm", "23cm", "13cm", "9cm", "6cm", "3cm", "1.25cm", "6mm", "4mm", "2.5mm", "2mm", "1mm"]
# The lower and upper frequency bounds (in MHz) for each band in BANDS.
BANDS_RANGES = [(None, None), (0.136, 0.137), (0.472, 0.479), (0.501, 0.504), (1.8, 2.0), (3.5, 4.0), (5.102, 5.4065), (7.0, 7.3), (10.0, 10.15), (14.0, 14.35), (18.068, 18.168), (21.0, 21.45), (24.890, 24.99), (28.0, 29.7), (50.0, 54.0), (70.0, 71.0), (144.0, 148.0), (222.0, 225.0), (420.0, 450.0), (902.0, 928.0), (1240.0, 1300.0), (2300.0, 2450.0), (3300.0, 3500.0), (5650.0, 5925.0), (10000.0, 10500.0), (24000.0, 24250.0), (47000.0, 47200.0), (75500.0, 81000.0), (119980.0, 120020.0), (142000.0, 149000.0), (241000.0, 250000.0)]

PROPAGATION_MODES = ["", "AS", "AUE", "AUR", "BS", "ECH", "EME", "ES", "F2", "FAI", "INTERNET", "ION", "IRL", "MS", "RPT", "RS", "SAT", "TEP", "TR"]

ADIF_VERSION = "3.0.4"


class ADIF:

    """ The ADIF class supplies methods for reading, parsing, and writing log files in the Amateur Data Interchange Format (ADIF).
    For more information, visit http://adif.org/ """

    def __init__(self):
        """ Initialise class for I/O of files using the Amateur Data Interchange Format (ADIF). """
        return

    def read(self, path):
        """ Read an ADIF file and parse it.

        :arg str path: The path to the ADIF file to read.
        :returns: A list of dictionaries (one dictionary per QSO), with each dictionary containing field-value pairs, e.g. {FREQ:145.500, BAND:2M, MODE:FM}. If the file cannot be read, the method returns None.
        :rtype: list
        :raises IOError: If the ADIF file does not exist or cannot be read (e.g. due to lack of read permissions).
        """
        logging.debug("Reading in ADIF file with path: %s..." % path)

        text = ""
        with open(path, mode="r", errors="replace") as f:
            text = f.read()

        records = self.parse_adi(text)

        if(records == []):
            logging.warning("No records found in the file. Empty file or wrong file type?")

        logging.info("Read %d QSOs from %s in ADIF format." % (len(records), path))
        return records

    def parse_adi(self, text):
        """ Parse some raw text (defined in the 'text' argument) for ADIF field data.

        :arg str text: The raw text from the ADIF file to parse.
        :returns: A list of dictionaries (one dictionary per QSO). Each dictionary contains the field-value pairs, e.g. {"FREQ": "145.500", "BAND": "2M", "MODE": "FM"}.
        :rtype: list
        """

        logging.debug("Parsing text from the ADIF file...")

        records = []

        # ADIF-related configuration options
        config = configparser.ConfigParser()
        have_config = (config.read(expanduser("~/.config/pyqso/preferences.ini")) != [])
        (section, option) = ("import_export", "merge_comment")
        if(have_config and config.has_option(section, option) and config.getboolean(section, option)):
            merge_comment = True
        else:
            merge_comment = False

        # Separate the text at the <eor> or <eoh> markers.
        tokens = re.split("(<eor>|<eoh>)", text, flags=re.IGNORECASE)
        tokens.pop()  # Anything after the final <eor> marker should be ignored.

        # The header might tell us the number of records, but let's not assume
        # this and simply ignore it instead (if it exists).
        if(re.search("<eoh>", text, flags=re.IGNORECASE) is not None):
            # There is a header present, so let's ignore everything
            # up to and including the <eoh> marker.
            while len(tokens) > 0:
                t = tokens.pop(0)
                if(re.match("<eoh>", t, flags=re.IGNORECASE) is not None):
                    break

        n_eor = 0
        n_record = 0
        records = []
        pattern = re.compile("<(.*?):(\d*).*?>([^<]+)")

        for t in tokens:
            if(re.match("<eor>", t, flags=re.IGNORECASE) is not None):
                n_eor += 1
                continue
            else:
                n_record += 1
                # Each record will have field names and corresponding
                # data entries. Store this in a dictionary.
                # Note: This is based on the code written by OK4BX.
                # (http://web.bxhome.org/blog/ok4bx/2012/05/adif-parser-python)
                fields_and_data_dictionary = {}
                fields_and_data = pattern.findall(t)
                comment = None
                for fd in fields_and_data:
                    # Let's force all field names to be in upper case.
                    # This will help us later when comparing the field names
                    # against the available field names in the ADIF specification.
                    field_name = fd[0].upper()
                    # Only read in the number of characters specified by the data length.
                    field_data = fd[2][:int(fd[1])]

                    # Combo boxes are used later on and these are case sensitive,
                    # so adjust the field data accordingly.
                    if(field_name == "BAND"):
                        field_data = field_data.lower()
                    elif(field_name == "CALL" or field_name == "MODE" or field_name == "SUBMODE"):
                        field_data = field_data.upper()
                    elif(field_name == "COMMENT"):
                        # Keep a copy of the COMMENT field data, in case we want to merge
                        # it with the NOTES field.
                        comment = field_data
                    if(field_name in AVAILABLE_FIELD_NAMES_ORDERED):
                        field_data_type = AVAILABLE_FIELD_NAMES_TYPES[field_name]
                        if(self.is_valid(field_name, field_data, field_data_type)):
                            # Only add the field if it is a standard ADIF field and it holds valid data.
                            fields_and_data_dictionary[field_name] = field_data

                # Merge the COMMENT field with the NOTES field, if desired and applicable.
                if(merge_comment):
                    if("NOTES" in list(fields_and_data_dictionary.keys()) and comment):
                        logging.debug("Merging COMMENT field with NOTES field...")
                        fields_and_data_dictionary["NOTES"] += "\n" + comment
                        logging.debug("Merged fields.")
                    elif(comment):
                        # Create the NOTES entry, but only store the contents of the COMMENT field.
                        logging.debug("The COMMENT field is present, but not the NOTES field. The NOTES field will be created and will only hold the COMMENT.")
                        fields_and_data_dictionary["NOTES"] = comment
                    else:
                        pass
                records.append(fields_and_data_dictionary)

        assert n_eor == n_record

        logging.debug("Finished parsing text.")

        return records

    def write(self, records, path):
        """ Write an ADIF file containing all the QSOs in the 'records' list.

        :arg list records: The list of QSO records to write.
        :arg str path: The desired path of the ADIF file to write to.
        :returns: None
        :raises IOError: If the ADIF file cannot be written (e.g. due to lack of write permissions).
        """

        logging.debug("Writing records to an ADIF file...")

        with open(path, mode="w", errors="replace") as f:  # Open file for writing

            # First write a header containing program version, number of records, etc.
            dt = datetime.now()

            f.write("""Amateur radio log file. Generated on %s. Contains %d record(s).

<adif_ver:%d>%s
<programid:5>PyQSO
<programversion:5>1.1.0
<eoh>\n""" % (dt, len(records), len(str(ADIF_VERSION)), ADIF_VERSION))

            # Then write each record to the file.
            for r in records:
                for field_name in AVAILABLE_FIELD_NAMES_ORDERED:
                    if(not(field_name.lower() in list(r.keys()) or field_name.upper() in list(r.keys()))):
                        # If the field_name does not exist in the record, then skip past it.
                        # Only write out the fields that exist and that have some data in them.
                        continue
                    else:
                        if((r[field_name] != "NULL") and (r[field_name] != "")):
                            f.write("<%s:%d>%s\n" % (field_name.lower(), len(r[field_name]), r[field_name]))
                f.write("<eor>\n")

            logging.debug("Finished writing records to the ADIF file.")
            f.close()

            logging.info("Wrote %d QSOs to %s in ADIF format." % (len(records), path))

        return

    def is_valid(self, field_name, data, data_type):
        """ Validate the data in a field with respect to the ADIF specification.

        :arg str field_name: The name of the ADIF field.
        :arg str data: The data of the ADIF field to validate.
        :arg str data_type: The type of data to be validated. See http://www.adif.org/304/ADIF_304.htm#Data_Types for the full list with descriptions.
        :returns: True or False to indicate whether the data is valid or not.
        :rtype: bool
        """

        logging.debug("Validating the following data in field '%s': %s" % (field_name, data))

        # Allow an empty string or None, in case the user doesn't want
        # to fill in this field.
        if(not data):
            return True

        if(data_type == "N"):
            # Allow a decimal point before and/or after any numbers,
            # but don't allow a decimal point on its own.
            m = re.match(r"-?(([0-9]+\.?[0-9]*)|([0-9]*\.?[0-9]+))", data)
            if(m is None):
                # Did not match anything.
                return False
            else:
                # Make sure we match the whole string,
                # otherwise there may be an invalid character after the match.
                return (m.group(0) == data)

        elif(data_type == "B"):
            # Boolean
            m = re.match(r"(Y|N)", data)
            if(m is None):
                return False
            else:
                return (m.group(0) == data)

        elif(data_type == "D"):
            # Date
            pattern = re.compile(r"([0-9]{4})")
            m_year = pattern.match(data, 0)
            if((m_year is None) or (int(m_year.group(0)) < 1930)):
                # Did not match anything.
                return False
            else:
                pattern = re.compile(r"([0-9]{2})")
                m_month = pattern.match(data, 4)
                if((m_month is None) or int(m_month.group(0)) > 12 or int(m_month.group(0)) < 1):
                    # Did not match anything.
                    return False
                else:
                    pattern = re.compile(r"([0-9]{2})")
                    m_day = pattern.match(data, 6)
                    days_in_month = calendar.monthrange(int(m_year.group(0)), int(m_month.group(0)))
                    if((m_day is None) or int(m_day.group(0)) > days_in_month[1] or int(m_day.group(0)) < 1):
                        # Did not match anything.
                        return False
                    else:
                        # Make sure we match the whole string,
                        # otherwise there may be an invalid character after the match.
                        return (len(data) == 8)

        elif(data_type == "T"):
            # Time
            pattern = re.compile(r"([0-9]{2})")
            m_hour = pattern.match(data, 0)
            if((m_hour is None) or (int(m_hour.group(0)) < 0) or (int(m_hour.group(0)) > 23)):
                # Did not match anything.
                return False
            else:
                pattern = re.compile(r"([0-9]{2})")
                m_minutes = pattern.match(data, 2)
                if((m_minutes is None) or int(m_minutes.group(0)) < 0 or int(m_minutes.group(0)) > 59):
                    # Did not match anything.
                    return False
                else:
                    if(len(data) == 4):
                        # HHMM format
                        return True
                    pattern = re.compile(r"([0-9]{2})")
                    m_seconds = pattern.match(data, 4)
                    if((m_seconds is None) or int(m_seconds.group(0)) < 0 or int(m_seconds.group(0)) > 59):
                        # Did not match anything.
                        return False
                    else:
                        # Make sure we match the whole string,
                        # otherwise there may be an invalid character after the match.
                        return (len(data) == 6)  # HHMMSS format

        # FIXME: Need to make sure that the "S" and "M" data types accept ASCII-only characters
        # in the range 32-126 inclusive.
        elif(data_type == "S"):
            # String
            m = re.match(r"(.+)", data)
            if(m is None):
                return False
            else:
                return (m.group(0) == data)

        elif(data_type == "I"):
            # IntlString
            m = re.match(r"(.+)", data, re.UNICODE)
            if(m is None):
                return False
            else:
                return (m.group(0) == data)

        elif(data_type == "G"):
            # IntlMultilineString
            m = re.match(r"(.+(\r\n)*.*)", data, re.UNICODE)
            if(m is None):
                return False
            else:
                return (m.group(0) == data)

        elif(data_type == "M"):
            # MultilineString
            # m = re.match(r"(.+(\r\n)*.*)", data)
            # if(m is None):
            #   return False
            # else:
            #   return (m.group(0) == data)
            return True

        elif(data_type == "L"):
            # Location
            pattern = re.compile(r"([EWNS]{1})", re.IGNORECASE)
            m_directional = pattern.match(data, 0)
            if(m_directional is None):
                # Did not match anything.
                return False
            else:
                pattern = re.compile(r"([0-9]{3})")
                m_degrees = pattern.match(data, 1)
                if((m_degrees is None) or int(m_degrees.group(0)) < 0 or int(m_degrees.group(0)) > 180):
                    # Did not match anything.
                    return False
                else:
                    pattern = re.compile(r"([0-9]{2}\.[0-9]{3})")
                    m_minutes = pattern.match(data, 4)
                    if((m_minutes is None) or float(m_minutes.group(0)) < 0 or float(m_minutes.group(0)) > 59.999):
                        # Did not match anything.
                        return False
                    else:
                        # Make sure we match the whole string,
                        # otherwise there may be an invalid character after the match.
                        return (len(data) == 10)

        elif(data_type == "E" or data_type == "A"):
            # Enumeration, AwardList.
            if(field_name == "MODE"):
                return (data in list(MODES.keys()))
            elif(field_name == "SUBMODE"):
                submodes = [submode for mode in list(MODES.keys()) for submode in MODES[mode]]
                return (data in submodes)
            elif(field_name == "BAND"):
                return (data in BANDS)
            else:
                return True

        else:
            return True



#!/usr/bin/env python3

#    Copyright (C) 2017 Christian Thomas Jacobs.

#    This file is part of PyQSO.

#    PyQSO is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyQSO is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyQSO.  If not, see <http://www.gnu.org/licenses/>.

import unittest
import os
from pyqso.adif import *


class TestADIF(unittest.TestCase):

    """ The unit tests for the ADIF class. """

    def setUp(self):
        """ Set up the ADIF object needed for the unit tests. """
        self.adif = ADIF()

    def test_read(self):
        """ Check that a single ADIF record can be read and parsed correctly. """
        path = os.path.join(os.path.realpath(os.path.dirname(__file__)), "res", "ADIF.test_read.adi")
        records = self.adif.read(path)
        expected_records = [{'TIME_ON': '1955', 'BAND': '40m', 'CALL': 'TEST', 'MODE': 'CW', 'QSO_DATE': '20130322'}]
        print("Imported records: ", records)
        print("Expected records: ", expected_records)
        assert(len(records) == 1)
        assert(len(list(records[0].keys())) == len(list(expected_records[0].keys())))
        assert(records == expected_records)

    def test_read_multiple(self):
        """ Check that multiple ADIF records can be read and parsed correctly. """
        path = os.path.join(os.path.realpath(os.path.dirname(__file__)), "res", "ADIF.test_read_multiple.adi")
        records = self.adif.read(path)
        expected_records = [{'TIME_ON': '1955', 'BAND': '40m', 'CALL': 'TEST', 'MODE': 'CW', 'QSO_DATE': '20130322'}, {'TIME_ON': '0820', 'BAND': '20m', 'CALL': 'TEST2ABC', 'MODE': 'SSB', 'QSO_DATE': '20150227'}, {'TIME_ON': '0832', 'BAND': '2m', 'CALL': 'HELLO', 'MODE': 'FM', 'QSO_DATE': '20150227'}]
        print("Imported records: ", records)
        print("Expected records: ", expected_records)
        assert(len(records) == 3)
        for i in range(len(expected_records)):
            assert(len(list(records[i].keys())) == len(list(expected_records[i].keys())))
        assert(records == expected_records)

    def test_read_alphabet(self):
        """ Check that none of the letters of the alphabet are ignored during parsing. """
        path = os.path.join(os.path.realpath(os.path.dirname(__file__)), "res", "ADIF.test_read_alphabet.adi")
        records = self.adif.read(path)
        expected_records = [{'CALL': 'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ'}]
        print("Imported records: ", records)
        print("Expected records: ", expected_records)
        assert(len(records) == 1)
        assert(len(list(records[0].keys())) == len(list(expected_records[0].keys())))
        assert(records == expected_records)

    def test_read_capitalisation(self):
        """ Check that the CALL field is capitalised correctly. """
        path = os.path.join(os.path.realpath(os.path.dirname(__file__)), "res", "ADIF.test_read_capitalisation.adi")
        records = self.adif.read(path)
        expected_records = [{'CALL': 'TEST'}]
        print("Imported records: ", records)
        print("Expected records: ", expected_records)
        assert(len(records) == 1)
        assert(len(list(records[0].keys())) == len(list(expected_records[0].keys())))
        assert(records == expected_records)

    def test_read_header_only(self):
        """ Check that no records are read in if the ADIF file only contains header information. """
        path = os.path.join(os.path.realpath(os.path.dirname(__file__)), "res", "ADIF.test_read_header_only.adi")
        records = self.adif.read(path)
        expected_records = []
        print("Imported records: ", records)
        print("Expected records: ", expected_records)
        assert(len(records) == 0)
        assert(records == expected_records)

    def test_read_no_header(self):
        """ Check that an ADIF file can be parsed with no header information. """
        path = os.path.join(os.path.realpath(os.path.dirname(__file__)), "res", "ADIF.test_read_no_header.adi")
        records = self.adif.read(path)
        expected_records = [{'TIME_ON': '1955', 'BAND': '40m', 'CALL': 'TEST', 'MODE': 'CW', 'QSO_DATE': '20130322'}]
        print("Imported records: ", records)
        print("Expected records: ", expected_records)
        assert(len(records) == 1)
        assert(len(list(records[0].keys())) == len(list(expected_records[0].keys())))
        assert(records == expected_records)

    def test_write(self):
        """ Check that records can be written to an ADIF file correctly. """
        records = [{"CALL": "TEST123", "QSO_DATE": "20120402", "TIME_ON": "1234", "FREQ": "145.500", "BAND": "2m", "MODE": "FM", "RST_SENT": "59", "RST_RCVD": "59"},
                   {"CALL": "TEST123", "QSO_DATE": "20130312", "TIME_ON": "0101", "FREQ": "145.750", "BAND": "2m", "MODE": "FM"}]
        self.adif.write(records, "ADIF.test_write.adi")

        f = open("ADIF.test_write.adi", 'r')
        text = f.read()
        print("File 'ADIF.test_write.adi' contains the following text:", text)
        assert("""
<adif_ver:5>3.0.4
<programid:5>PyQSO
<programversion:5>1.1.0
<eoh>
<call:7>TEST123
<qso_date:8>20120402
<time_on:4>1234
<freq:7>145.500
<band:2>2m
<mode:2>FM
<rst_sent:2>59
<rst_rcvd:2>59
<eor>
<call:7>TEST123
<qso_date:8>20130312
<time_on:4>0101
<freq:7>145.750
<band:2>2m
<mode:2>FM
<eor>
""" in text)  # Ignore the header line here, since it contains the date and time the ADIF file was written, which will change each time 'make unittest' is run.
        f.close()

    def test_write_sqlite3_Row(self):
        """ Check that records can be written to an ADIF file from a test database file. """
        import sqlite3
        import os.path
        path_to_test_database = os.path.join(os.path.realpath(os.path.dirname(__file__)), "res", "test.db")
        self.connection = sqlite3.connect(path_to_test_database)
        self.connection.row_factory = sqlite3.Row

        c = self.connection.cursor()
        c.execute("SELECT * FROM test")
        records = c.fetchall()
        print(records)

        self.adif.write(records, "ADIF.test_write_sqlite3_Row.adi")

        f = open("ADIF.test_write_sqlite3_Row.adi", 'r')
        text = f.read()
        print("File 'ADIF.test_write_sqlite3_Row.adi' contains the following text:", text)
        assert("""
<adif_ver:5>3.0.4
<programid:5>PyQSO
<programversion:5>1.1.0
<eoh>
<call:7>TEST123
<qso_date:8>20120402
<time_on:4>1234
<freq:7>145.500
<band:2>2m
<mode:2>FM
<rst_sent:2>59
<rst_rcvd:2>59
<eor>
<call:7>TEST456
<qso_date:8>20130312
<time_on:4>0101
<freq:7>145.750
<band:2>2m
<mode:2>FM
<eor>
""" in text)  # Ignore the header line here, since it contains the date and time the ADIF file was written, which will change each time 'make unittest' is run.
        f.close()

        self.connection.close()

    def test_is_valid(self):
        """ Check that ADIF field validation is working correctly for different data types. """

        assert(self.adif.is_valid("CALL", "TEST123", "S"))
        assert(self.adif.is_valid("CALL", "F/MYCALL123MYCALL", "S"))

        assert(self.adif.is_valid("QSO_DATE", "20120402", "D"))
        assert(not self.adif.is_valid("QSO_DATE", "19000101", "D"))

        assert(self.adif.is_valid("TIME_ON", "0000", "T"))
        assert(self.adif.is_valid("TIME_ON", "235959", "T"))
        assert(self.adif.is_valid("TIME_ON", "1230", "T"))
        assert(self.adif.is_valid("TIME_ON", "155329", "T"))
        assert(not self.adif.is_valid("TIME_ON", "2500", "T"))

        assert(self.adif.is_valid("TX_PWR", "5", "N"))
        assert(self.adif.is_valid("FREQ", "145.550", "N"))

        assert(self.adif.is_valid("NOTES", "TEST123\nHELLO_WORLD", "M"))

        assert(self.adif.is_valid("MODE", "FM", "E"))
        assert(self.adif.is_valid("SUBMODE", "LSB", "E"))

if(__name__ == '__main__'):
    unittest.main()



#!/usr/bin/env python3

#    Copyright (C) 2013-2017 Christian Thomas Jacobs.

#    This file is part of PyQSO.

#    PyQSO is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyQSO is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyQSO.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk
import logging
import sqlite3 as sqlite


class Awards:

    """ A tool for tracking progress towards an award. Currently this only supports the DXCC award.
    For more information visit http://www.arrl.org/dxcc """

    def __init__(self, application):
        """ Set up a table for progress tracking purposes.

        :arg application: The PyQSO application containing the main Gtk window, etc.
        """
        # TODO: Add more awards. This only considers the DXCC award for now.
        logging.debug("Setting up awards table...")

        self.application = application
        self.builder = self.application.builder

        self.bands = ["70cm", "2m", "6m", "10m", "12m", "15m", "17m", "20m", "30m", "40m", "80m", "160m"]
        self.modes = ["Phone", "CW", "Digital", "Mixed"]

        data_types = [str] + [int]*len(self.bands)
        self.awards = Gtk.ListStore(*data_types)

        # The main table for the awards.
        self.treeview = Gtk.TreeView(model=self.awards)
        # A separate, empty column just for the mode names.
        renderer = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn("Modes", renderer, text=0)
        column.set_clickable(False)
        self.treeview.append_column(column)
        # Now for all the bands...
        logging.debug("Initialising the columns in the awards table.")
        for i in range(0, len(self.bands)):
            renderer = Gtk.CellRendererText()
            column = Gtk.TreeViewColumn(self.bands[i], renderer, text=i+1)
            column.set_min_width(40)
            column.set_clickable(False)
            self.treeview.append_column(column)

        # Show the table in the Awards tab.
        self.builder.get_object("awards").add(self.treeview)
        self.builder.get_object("awards").show_all()

        logging.debug("Awards table set up successfully.")

        self.count(self.application.logbook)

        return

    def count(self, logbook):
        """ Update the table for progress tracking.

        :arg logbook: The logbook containing logs which in turn contain QSOs.
        :returns: A list of lists containing the QSO counts for different modes and bands.
        :rtype: list
        """

        logging.debug("Counting the band/mode combinations for the awards table...")

        # Wipe everything and start again.
        self.awards.clear()

        # For each mode, add a new list for holding the totals, and initialise the values to zero.
        count = []
        for i in range(0, len(self.bands)):
            count.append([0]*len(self.bands))

        for log in logbook.logs:
            try:
                records = log.records
                for r in records:
                    if(r["BAND"] is not None and r["MODE"] is not None):
                        if(r["BAND"].lower() in self.bands and r["MODE"] != ""):
                            band = self.bands.index(r["BAND"].lower())
                            # Phone modes
                            if(r["MODE"].upper() in ["FM", "AM", "SSB", "SSTV"]):
                                count[0][band] += 1
                            elif(r["MODE"].upper() == "CW"):
                                count[1][band] += 1
                            else:
                                # FIXME: This assumes that all the other modes in the ADIF list are digital modes. Is this the case?
                                count[2][band] += 1
                            count[3][band] += 1  # Keep the total of each column in the "Mixed" mode.

            except sqlite.Error as e:
                logging.error("Could not update the awards table for '%s' because of a database error." % log.name)
                logging.exception(e)

        # Insert the rows containing the totals.
        for i in range(0, len(self.modes)):
            self.awards.append([self.modes[i]] + count[i])

        logging.debug("Awards table updated.")
        return count



#!/usr/bin/env python3

#    Copyright (C) 2017 Christian Thomas Jacobs.

#    This file is part of PyQSO.

#    PyQSO is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyQSO is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyQSO.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk
import unittest
try:
    import unittest.mock as mock
except ImportError:
    import mock
import os
from pyqso.printer import *


class TestPrinter(unittest.TestCase):

    """ The unit tests for the Printer class. """

    def setUp(self):
        """ Set up the Printer object. """
        PyQSO = mock.MagicMock()
        self.printer = Printer(application=PyQSO())
        self.printer.application.window = Gtk.Window()

    def test_print_records(self):
        """ Check that a list of records can be printed to a PDF file. """
        self.printer.action = Gtk.PrintOperationAction.EXPORT
        pdf = "Printer.test_print_records.pdf"
        self.printer.operation.set_export_filename(pdf)
        records = [{"id": 1, "CALL": "MYCALL", "QSO_DATE": "24062017", "TIME_ON": "1519", "FREQ": "145.550", "MODE": "FM", "RST_SENT": "59", "RST_RCVD": "57"}]
        result = self.printer.print_records(records)
        assert(result != Gtk.PrintOperationResult.ERROR)
        assert(result == Gtk.PrintOperationResult.APPLY)
        assert(os.path.exists(pdf))

if(__name__ == '__main__'):
    unittest.main()



#!/usr/bin/env python3

#    Copyright (C) 2017 Christian Thomas Jacobs.

#    This file is part of PyQSO.

#    PyQSO is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyQSO is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyQSO.  If not, see <http://www.gnu.org/licenses/>.

import unittest
try:
    import unittest.mock as mock
except ImportError:
    import mock
import os
from shutil import copyfile
from pyqso.logbook import *


class TestLogbook(unittest.TestCase):

    """ The unit tests for the Logbook class. """

    @mock.patch('pyqso.logbook.Logbook.filter_by_callsign')
    def setUp(self, mock_filter_by_callsign):
        """ Set up the Logbook object and connection to the test database needed for the unit tests. """

        self.logbook = Logbook(application=mock.MagicMock())

        # Open the test database file.
        path_to_test_database = os.path.join(os.path.realpath(os.path.dirname(__file__)), "res", "test.db")
        opened = self.logbook.open(path=path_to_test_database)
        assert(opened)
        assert(self.logbook.connection is not None)

        # Check that the logs have been retrieved.
        assert(len(self.logbook.logs) == 2)
        assert(self.logbook.logs[0].name == "test")
        assert(self.logbook.logs[1].name == "test2")

    def tearDown(self):
        """ Close the logbook and disconnect from the test database. """
        self.logbook.notebook.get_n_pages.return_value = 0
        closed = self.logbook.close()
        assert(closed)

    def test_db_disconnect(self):
        """ Check that the logbook can disconnect from the database. """
        disconnected = self.logbook.db_disconnect()
        assert(disconnected)
        # Attempt to disconnect again. This shouldn't do anything.
        disconnected = self.logbook.db_disconnect()
        assert(disconnected)

    @mock.patch('pyqso.auxiliary_dialogs.handle_gtk_dialog')
    @mock.patch('gi.repository.Gtk.FileChooserDialog')
    def test_new(self, mock_FileChooserDialog, mock_handle_gtk_dialog):
        """ Check that a new logbook can be created. """
        mock_FileChooserDialog().run.return_value = Gtk.ResponseType.OK
        mock_FileChooserDialog().get_filename.return_value = "Logbook.test_new.db"
        self.logbook.new()
        assert(os.path.isfile("Logbook.test_new.db"))

    @mock.patch('pyqso.auxiliary_dialogs.handle_gtk_dialog')
    def test_open_invalid_logbook(self, mock_handle_gtk_dialog):
        """ Open an invalid database file (comprising only one line of plain text) and check that an error occurs. """
        invalid_logbook = Logbook(application=mock.MagicMock())
        path_to_invalid_database = os.path.join(os.path.realpath(os.path.dirname(__file__)), "res", "invalid.db")
        opened = invalid_logbook.open(path=path_to_invalid_database)
        assert(not opened)
        assert(not invalid_logbook.logs)

    @mock.patch('pyqso.logbook.Logbook.render_log')
    @mock.patch('pyqso.auxiliary_dialogs.handle_gtk_dialog')
    @mock.patch('pyqso.logbook.LogNameDialog')
    def test_new_log(self, mock_LogNameDialog, mock_handle_gtk_dialog, mock_render_log):
        """ Create an empty logbook file, open it, and check that a new log can successfully be added. """
        # Create a copy of the test database just for use in this particular test, since the contents will need to be modified.
        path_to_test_database = os.path.join(os.path.realpath(os.path.dirname(__file__)), "res", "test.db")
        destination = "Logbook.test_new_log.db"
        copyfile(path_to_test_database, destination)
        opened = self.logbook.open(path=destination)
        assert(opened)
        mock_LogNameDialog().dialog.run.return_value = Gtk.ResponseType.OK
        mock_LogNameDialog().name = "my_new_log"
        self.logbook.new_log()
        assert(len(self.logbook.logs) == 3)
        assert(self.logbook.logs[-1].name == "my_new_log")

    def test_log_name_exists(self):
        """ Check that only the log called 'test' exists. """
        assert(self.logbook.log_name_exists("test"))  # Log 'test' exists.
        assert(not self.logbook.log_name_exists("hello"))  # Log 'hello' should not exist.

    def test_log_count(self):
        """ Check the log count. """
        assert(self.logbook.log_count == 2)

    def test_record_count(self):
        """ Check the total number of records over all the logs in the logbook. """
        assert(self.logbook.record_count == 7)

    def test_filter_by_callsign(self):
        """ Check that callsigns are filtered correctly. """

        # Consider only the first record of the first log.
        model = self.logbook.logs[0]
        path = Gtk.TreePath(0)
        iter = model.get_iter(path)

        self.logbook.application.toolbar.filter_source.get_text.return_value = ""
        present = self.logbook.filter_by_callsign(model, iter, data=None)
        assert(present)  # Show all the callsigns.

        self.logbook.application.toolbar.filter_source.get_text.return_value = "TEST123"
        present = self.logbook.filter_by_callsign(model, iter, data=None)
        assert(present)  # "TEST123" is present.

        self.logbook.application.toolbar.filter_source.get_text.return_value = "TEST"
        present = self.logbook.filter_by_callsign(model, iter, data=None)
        assert(present)  # "TEST" is present in "TEST123"

        self.logbook.application.toolbar.filter_source.get_text.return_value = "HELLOWORLD"
        present = self.logbook.filter_by_callsign(model, iter, data=None)
        assert(not present)  # "HELLOWORLD" is not present in "TEST123"

    def test_get_log_index(self):
        """ Check that a log's index can be resolved using the log's name. """
        assert(self.logbook.get_log_index(name="test") == 0)
        assert(self.logbook.get_log_index(name="test2") == 1)
        assert(self.logbook.get_log_index(name="helloworld") is None)

if(__name__ == '__main__'):
    unittest.main()



#!/usr/bin/env python3

#    Copyright (C) 2013-2017 Christian Thomas Jacobs.

#    This file is part of PyQSO.

#    PyQSO is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyQSO is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyQSO.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GObject, Gdk
import logging
import telnetlib
try:
    import configparser
except ImportError:
    import ConfigParser as configparser
import os.path

from pyqso.telnet_connection_dialog import TelnetConnectionDialog
from pyqso.auxiliary_dialogs import error

BOOKMARKS_FILE = os.path.expanduser('~/.config/pyqso/bookmarks.ini')


class DXCluster:

    """ A tool for connecting to a DX cluster (specifically Telnet-based DX clusters). """

    def __init__(self, application):
        """ Set up the DX cluster, and set up a timer so that PyQSO can retrieve new data from the Telnet server every few seconds.

        :arg application: The PyQSO application containing the main Gtk window, etc.
        """

        logging.debug("Setting up the DX cluster...")

        self.application = application
        self.builder = self.application.builder
        self.connection = None

        # Connect signals.
        self.builder.get_object("mitem_new").connect("activate", self.new_server)
        self.builder.get_object("mitem_disconnect").connect("activate", self.telnet_disconnect)
        self.builder.get_object("send").connect("clicked", self.telnet_send_command)
        self.builder.get_object("command").connect("key-press-event", self.on_command_key_press)

        # Get the text renderer and its buffer.
        self.renderer = self.builder.get_object("renderer")
        self.buffer = self.renderer.get_buffer()

        # Items whose sensitivity may change.
        self.items = {}
        self.items["CONNECT"] = self.builder.get_object("mitem_connect")
        self.items["DISCONNECT"] = self.builder.get_object("mitem_disconnect")
        self.items["SEND"] = self.builder.get_object("send")
        self.set_items_sensitive(True)

        self.populate_bookmarks()

        logging.debug("DX cluster ready!")

        return

    def on_command_key_press(self, widget, event, data=None):
        """ If the Return key is pressed when the focus is on the command box, then send whatever command the user has entered. """
        if(event.keyval == Gdk.KEY_Return):
            self.telnet_send_command()
        return

    def new_server(self, widget=None):
        """ Get Telnet server host and login details specified in the Gtk.Entry boxes in the Telnet connection dialog and attempt a connection. """

        # Get connection details.
        tcd = TelnetConnectionDialog(self.application)
        response = tcd.dialog.run()
        if(response == Gtk.ResponseType.OK):
            host = tcd.host
            port = tcd.port
            username = tcd.username
            password = tcd.password
            bookmark = tcd.bookmark
            tcd.dialog.destroy()

            # Handle empty hostname.
            if(not host):
                logging.error("No hostname specified.")
                return

            # Handle empty port number.
            if(not port):
                logging.warning("No port specified. Assuming default port 23...")
                port = 23
            else:
                try:
                    # Cast port into an int.
                    port = int(port)
                except ValueError as e:
                    logging.error("Could not cast the DX cluster's port information to an integer.")
                    logging.exception(e)
                    return

            # Save the server details in a new bookmark, if desired.
            if(bookmark):
                try:
                    config = configparser.ConfigParser()
                    config.read(BOOKMARKS_FILE)

                    # Use the host name as the bookmark's identifier.
                    if(username):
                        bookmark_identifier = "%s@%s:%d" % (username, host, port)
                    else:
                        bookmark_identifier = "%s:%d" % (host, port)
                    logging.debug("Using %s as the bookmark identifier." % bookmark_identifier)

                    # Add bookmark.
                    try:
                        config.add_section(bookmark_identifier)
                    except configparser.DuplicateSectionError:
                        # If the hostname already exists, assume the user wants to update the port number, username and/or password.
                        logging.warning("Bookmark '%s' already exists. Over-writing existing details..." % (bookmark_identifier))
                    config.set(bookmark_identifier, "host", host)
                    config.set(bookmark_identifier, "port", str(port))
                    config.set(bookmark_identifier, "username", username)
                    config.set(bookmark_identifier, "password", password)

                    # Write the bookmarks to file.
                    if not os.path.exists(os.path.expanduser('~/.config/pyqso')):
                        os.makedirs(os.path.expanduser('~/.config/pyqso'))
                    with open(BOOKMARKS_FILE, 'w') as f:
                        config.write(f)

                    self.populate_bookmarks()

                except IOError:
                    # Maybe the bookmarks file could not be written to?
                    logging.error("Bookmark could not be saved. Check bookmarks file permissions? Going ahead with the server connection anyway...")

            # Attempt a connection with the server.
            self.telnet_connect(host, port, username, password)

        else:
            tcd.dialog.destroy()
        return

    def populate_bookmarks(self):
        """ Populate the list of bookmarked Telnet servers in the menu. """

        # Get the bookmarks submenu.
        subm_bookmarks = self.builder.get_object("subm_bookmarks")

        config = configparser.ConfigParser()
        have_config = (config.read(BOOKMARKS_FILE) != [])

        if(have_config):
            try:
                # Clear the menu of all current bookmarks.
                for i in subm_bookmarks.get_children():
                    subm_bookmarks.remove(i)

                # Add all bookmarks in the config file.
                for bookmark in config.sections():
                    mitem = Gtk.MenuItem(label=bookmark)
                    mitem.connect("activate", self.bookmarked_server, bookmark)
                    subm_bookmarks.append(mitem)

            except Exception as e:
                logging.error("An error occurred whilst populating the DX cluster bookmarks menu.")
                logging.exception(e)

            self.builder.get_object("dx_cluster").show_all()  # Need to do this to update the bookmarks list in the menu.

        return

    def bookmarked_server(self, widget, name):
        """ Get Telnet server host and login details from an existing bookmark and attempt a connection.

        :arg str name: The name of the bookmark. This is the same as the server's hostname.
        """

        config = configparser.ConfigParser()
        have_config = (config.read(BOOKMARKS_FILE) != [])
        try:
            if(not have_config):
                raise IOError("The bookmark's details could not be loaded.")

            host = config.get(name, "host")
            port = int(config.get(name, "port"))
            username = config.get(name, "username")
            password = config.get(name, "password")
            self.telnet_connect(host, port, username, password)

        except ValueError as e:
            # This exception may occur when casting the port (which is a str) to an int.
            logging.exception(e)
        except IOError as e:
            logging.exception(e)
        except Exception as e:
            logging.error("Could not connect to Telnet server '%s'." % name)
            logging.exception(e)

        return

    def telnet_connect(self, host, port=23, username=None, password=None):
        """ Connect to a user-specified Telnet server.

        :arg str host: The Telnet server's hostname.
        :arg int port: The Telnet server's port number. If no port is specified, the default Telnet server port of 23 will be used.
        :arg str username: The user's username. This is an optional argument.
        :arg str password: The user's password. This is an optional argument.
        """

        # Handle empty host/port string (or the case where host/port are None).
        if(not host):
            error(parent=self.application.window, message="Unable to connect to a DX cluster because no hostname was specified.")
            return
        if(not port):
            logging.warning("No port specified. Assuming default port 23...")
            port = 23  # Use the default Telnet port.

        try:
            logging.debug("Attempting connection to Telnet server %s:%d..." % (host, port))
            self.connection = telnetlib.Telnet(host, port)
            assert(self.connection)

            if(username):
                self.connection.read_until("login: ".encode())
                self.connection.write((username + "\n").encode())
            if(password):
                self.connection.read_until("password: ".encode())
                self.connection.write((password + "\n").encode())
        except Exception as e:
            message = "Could not create a connection to the Telnet server %s:%d. Check connection to the internets? Check connection details?" % (host, port)
            error(parent=self.application.window, message=message)
            logging.exception(e)
            self.connection = None
            return

        logging.debug("Connection to %s:%d established." % (host, port))

        self.set_items_sensitive(False)

        self.check_io_event = GObject.timeout_add(1000, self.on_telnet_io)

        return

    def telnet_disconnect(self, widget=None):
        """ Disconnect from a Telnet server and remove the I/O timer. """
        if(self.connection):
            self.connection.close()
        self.buffer.set_text("")
        self.connection = None
        self.set_items_sensitive(True)

        # Stop checking for server output once disconnected.
        try:
            GObject.source_remove(self.check_io_event)
        except AttributeError:
            # This may happen if a connection hasn't yet been established.
            pass

        return

    def telnet_send_command(self, widget=None):
        """ Send the user-specified command in the Gtk.Entry box to the Telnet server (if PyQSO is connected to one). """
        if(self.connection):
            command = self.builder.get_object("command")
            self.connection.write((command.get_text() + "\n").encode())
            command.set_text("")
        return

    def on_telnet_io(self):
        """ Retrieve any new data from the Telnet server and print it out in the Gtk.TextView widget.

        :returns: Always returns True to satisfy the GObject timer.
        :rtype: bool
        """
        if(self.connection):
            text = self.connection.read_very_eager()
            text = text.decode("ascii", "replace")  # Replace any characters that cannot be decoded with a replacement marker.
            try:
                text = text.replace("\u0007", "")  # Remove the BEL Unicode character from the end of the line
            except UnicodeDecodeError:
                pass

            # Allow auto-scrolling to the new text entry if the focus is already at
            # the very end of the Gtk.TextView. Otherwise, don't auto-scroll
            # in case the user is reading something further up.
            # Note: This is based on the code from http://forums.gentoo.org/viewtopic-t-445598-view-next.html
            end_iter = self.buffer.get_end_iter()
            end_mark = self.buffer.create_mark(None, end_iter)
            self.renderer.move_mark_onscreen(end_mark)
            at_end = self.buffer.get_iter_at_mark(end_mark).equal(end_iter)
            self.buffer.insert(end_iter, text)
            if(at_end):
                end_mark = self.buffer.create_mark(None, end_iter)
                self.renderer.scroll_mark_onscreen(end_mark)

        return True

    def set_items_sensitive(self, sensitive):
        """ Enable/disable the relevant buttons for connecting/disconnecting from a DX cluster, so that users cannot click the connect button if PyQSO is already connected.

        :arg bool sensitive: If True, enable the Connect button and disable the Disconnect button. If False, vice versa.
        """
        self.items["CONNECT"].set_sensitive(sensitive)
        self.items["DISCONNECT"].set_sensitive(not sensitive)
        self.items["SEND"].set_sensitive(not sensitive)
        return



#!/usr/bin/env python3

#    Copyright (C) 2013-2017 Christian Thomas Jacobs.

#    This file is part of PyQSO.

#    PyQSO is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyQSO is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyQSO.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk
import logging


def error(parent, message):
    """ Display an error message.

    :arg parent: The Gtk parent window/dialog.
    :arg str message: The message to display to the user.
    """
    logging.error(message)
    handle_gtk_dialog(parent, Gtk.MessageType.ERROR, message, "Error")


def info(parent, message):
    """ Display some information.

    :arg parent: The Gtk parent window/dialog.
    :arg str message: The message to display to the user.
    """
    logging.info(message)
    handle_gtk_dialog(parent, Gtk.MessageType.INFO, message, "Information")


def question(parent, message):
    """ Ask the user a question. The dialog comes with 'Yes' and 'No' response buttons.

    :arg parent: The Gtk parent window/dialog.
    :arg str message: The message to display to the user.
    :returns: The 'yes'/'no' response from the user.
    :rtype: Gtk.ResponseType
    """
    return handle_gtk_dialog(parent, Gtk.MessageType.QUESTION, message, "Question")


def handle_gtk_dialog(parent, msgtype, message, title):
    """
    Instantiate and present a dialog to the user.

    :arg parent: The Gtk parent window/dialog.
    :arg Gtk.MessageType msgtype: The type of message to present to the user (e.g. a question, or error message).
    :arg str message: The message to display in the dialog.
    :arg str title: The title to display at the top of the dialog.
    :returns: The response from the user, based on which button they pushed.
    :rtype: Gtk.ResponseType
    """
    bt = Gtk.ButtonsType
    buttons = bt.YES_NO if msgtype == Gtk.MessageType.QUESTION else bt.OK
    dialog = Gtk.MessageDialog(parent, Gtk.DialogFlags.DESTROY_WITH_PARENT,
                               msgtype, buttons, message, title=title)
    response = dialog.run()
    dialog.destroy()
    return response



#!/usr/bin/env python3

#    Copyright (C) 2017 Christian Thomas Jacobs.

#    This file is part of PyQSO.

#    PyQSO is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyQSO is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyQSO.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk
import unittest
from pyqso.compare import *


class TestCompare(unittest.TestCase):

    """ The unit tests for the comparison schemes. """

    def setUp(self):
        """ Set up the objects needed for the unit tests. """

        data_types = [int] + [str]*3
        self.model = Gtk.ListStore(*data_types)
        row1 = [0, "100", "20150323", "1433"]
        self.model.append(row1)
        row2 = [1, "5000", "20160423", "1432"]
        self.model.append(row2)
        row3 = [2, "5000", "20160423", "1433"]
        self.model.append(row3)
        row4 = [3, "25", "20160423", "1433"]
        self.model.append(row4)
        return

    def test_compare_default(self):
        """ Check the correctness of the default comparison scheme. """

        # Get the row iterables.
        path = Gtk.TreePath(0)
        iter1 = self.model.get_iter(path)
        iter2 = self.model.iter_next(iter1)
        iter3 = self.model.iter_next(iter2)
        iter4 = self.model.iter_next(iter3)

        # Compare values in the second column.
        column_index = 1
        result = compare_default(self.model, iter1, iter2, column_index)
        assert(result == -1)
        result = compare_default(self.model, iter2, iter3, column_index)
        assert(result == 0)
        result = compare_default(self.model, iter3, iter4, column_index)
        assert(result == 1)

    def test_compare_date_and_time(self):
        """ Check that dates in yyyymmdd format are compared correctly. """

        # Get the row iterables.
        path = Gtk.TreePath(0)
        iter1 = self.model.get_iter(path)
        iter2 = self.model.iter_next(iter1)
        iter3 = self.model.iter_next(iter2)
        iter4 = self.model.iter_next(iter3)

        # Compare values in the third (and fourth, if necessary) column.
        column_index = 2
        result = compare_date_and_time(self.model, iter1, iter2, [column_index, column_index+1])
        assert(result == -1)
        result = compare_date_and_time(self.model, iter2, iter3, [column_index, column_index+1])
        assert(result == -1)
        result = compare_date_and_time(self.model, iter3, iter4, [column_index, column_index+1])
        assert(result == 0)
        result = compare_date_and_time(self.model, iter4, iter1, [column_index, column_index+1])
        assert(result == 1)

if(__name__ == '__main__'):
    unittest.main()



#!/usr/bin/env python3

#    Copyright (C) 2017 Christian Thomas Jacobs.

#    This file is part of PyQSO.

#    PyQSO is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyQSO is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyQSO.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk
import logging
from os.path import basename, getmtime, expanduser, dirname, join, realpath
from datetime import datetime, date
try:
    import configparser
except ImportError:
    import ConfigParser as configparser
try:
    import matplotlib
    matplotlib.use('Agg')
    matplotlib.rcParams['font.size'] = 10.0
    from matplotlib.backends.backend_gtk3cairo import FigureCanvasGTK3Cairo as FigureCanvas
    from matplotlib.figure import Figure
    from matplotlib.dates import DateFormatter, MonthLocator
    have_matplotlib = True
except ImportError as e:
    logging.warning(e)
    logging.warning("Could not import matplotlib, so you will not be able to plot annual logbook statistics. Check that all the PyQSO dependencies are satisfied.")
    have_matplotlib = False


class Summary(object):

    def __init__(self, application):
        """ Create a summary page containing various statistics such as the number of logs in the logbook, the logbook's modification date, etc.

        :arg application: The PyQSO application containing the main Gtk window, etc.
        """

        self.application = application
        self.logbook = self.application.logbook
        self.builder = self.application.builder
        glade_file_path = join(realpath(dirname(__file__)), "res", "pyqso.glade")
        self.builder.add_objects_from_file(glade_file_path, ("summary_page",))
        self.summary_page = self.builder.get_object("summary_page")

        self.items = {}

        # Database name in large font at the top of the summary page
        self.builder.get_object("database_name").set_markup("<span size=\"x-large\">%s</span>" % basename(self.logbook.path))
        self.items["LOG_COUNT"] = self.builder.get_object("log_count")
        self.items["QSO_COUNT"] = self.builder.get_object("qso_count")
        self.items["DATE_MODIFIED"] = self.builder.get_object("date_modified")

        # Yearly statistics
        config = configparser.ConfigParser()
        have_config = (config.read(expanduser('~/.config/pyqso/preferences.ini')) != [])
        (section, option) = ("general", "show_yearly_statistics")
        if(have_config and config.has_option(section, option)):
            if(config.getboolean("general", "show_yearly_statistics") and have_matplotlib):
                hbox = Gtk.HBox()
                label = Gtk.Label(label="Display statistics for year: ", halign=Gtk.Align.START)
                hbox.pack_start(label, False, False, 6)
                year_select = Gtk.ComboBoxText()
                min_year, max_year = self.get_year_bounds()
                if min_year and max_year:
                    for year in range(max_year, min_year-1, -1):
                        year_select.append_text(str(year))
                year_select.append_text("")
                year_select.connect("changed", self.on_year_changed)
                hbox.pack_start(year_select, False, False, 6)
                self.summary_page.pack_start(hbox, False, False, 4)

                self.items["YEARLY_STATISTICS"] = Figure()
                canvas = FigureCanvas(self.items["YEARLY_STATISTICS"])
                canvas.set_size_request(800, 175)
                canvas.show()
                self.summary_page.pack_start(canvas, True, True, 0)

        # Summary tab label and icon.
        tab = Gtk.HBox(homogeneous=False, spacing=0)
        label = Gtk.Label(label="Summary  ")
        icon = Gtk.Image.new_from_icon_name(Gtk.STOCK_INDEX, Gtk.IconSize.MENU)
        tab.pack_start(label, False, False, 0)
        tab.pack_start(icon, False, False, 0)
        tab.show_all()

        self.logbook.notebook.insert_page(self.summary_page, tab, 0)  # Append as a new tab
        self.logbook.notebook.show_all()

        return

    def on_year_changed(self, combo):
        """ Re-plot the statistics for the year selected by the user. """

        # Clear figure
        self.items["YEARLY_STATISTICS"].clf()
        self.items["YEARLY_STATISTICS"].canvas.draw()

        # Get year to show statistics for.
        year = combo.get_active_text()
        try:
            year = int(year)
        except ValueError:
            # Empty year string.
            return

        # Number of contacts made each month
        contact_count_plot = self.items["YEARLY_STATISTICS"].add_subplot(121)
        contact_count = self.get_annual_contact_count(year)

        # x-axis formatting based on the date
        contact_count_plot.bar(list(contact_count.keys()), list(contact_count.values()), color="k", width=15, align="center")
        formatter = DateFormatter("%b")
        contact_count_plot.xaxis.set_major_formatter(formatter)
        month_locator = MonthLocator()
        contact_count_plot.xaxis.set_major_locator(month_locator)
        contact_count_plot.set_ylabel("Number of QSOs")

        # Set x-axis upper limit based on the current month.
        contact_count_plot.xaxis_date()
        contact_count_plot.set_xlim([date(year-1, 12, 16), date(year, 12, 15)])  # Make a bit of space either side of January and December of the selected year.

        # Pie chart of all the modes used.
        mode_count_plot = self.items["YEARLY_STATISTICS"].add_subplot(122)
        mode_count = self.get_annual_mode_count(year)
        (patches, texts, autotexts) = mode_count_plot.pie(list(mode_count.values()), labels=mode_count.keys(), autopct='%1.1f%%', shadow=False)
        for p in patches:
            # Make the patches partially transparent.
            p.set_alpha(0.75)
        mode_count_plot.set_title("Modes used")

        self.items["YEARLY_STATISTICS"].canvas.draw()

        return

    def get_year_bounds(self):
        """ Find the years of the oldest and newest QSOs across all logs in the logbook.

        :returns: The years of the oldest and newest QSOs. The tuple (None, None) is returned if no QSOs have been made or no QSO dates have been specified.
        :rtype: tuple
        """

        c = self.logbook.connection.cursor()
        max_years = []
        min_years = []
        for log in self.logbook.logs:
            query = "SELECT min(QSO_DATE), max(QSO_DATE) FROM %s" % (log.name)
            c.execute(query)
            years = c.fetchone()
            if years[0] and years[1]:
                min_years.append(int(years[0][:4]))
                max_years.append(int(years[1][:4]))

        if len(min_years) == 0 or len(max_years) == 0:
            return None, None
        else:
            # Return the min and max across all logs.
            return min(min_years), max(max_years)

    def get_annual_contact_count(self, year):
        """ Find the total number of contacts made in each month in the specified year.

        :arg int year: The year of interest.
        :returns: The total number of contacts made in each month of a given year.
        :rtype: dict
        """

        contact_count = {}
        c = self.logbook.connection.cursor()

        for log in self.logbook.logs:
            query = "SELECT QSO_DATE, count(QSO_DATE) FROM %s WHERE QSO_DATE >= %d0101 AND QSO_DATE < %d0101 GROUP by QSO_DATE" % (log.name, year, year+1)
            c.execute(query)
            xy = c.fetchall()

            for i in range(len(xy)):
                date_str = xy[i][0]
                y = int(date_str[0:4])
                m = int(date_str[4:6])
                date = datetime(y, m, 1)  # Collect all contacts together by month.
                if date in contact_count.keys():
                    contact_count[date] += xy[i][1]
                else:
                    contact_count[date] = xy[i][1]

        return contact_count

    def get_annual_mode_count(self, year):
        """ Find the total number of contacts made with each mode in a specified year.

        :arg int year: The year of interest.
        :returns: The total number of contacts made with each mode in a given year.
        :rtype: dict
        """

        mode_count = {}

        for log in self.logbook.logs:
            query = "SELECT MODE, count(MODE) FROM %s WHERE QSO_DATE >= %d0101 AND QSO_DATE < %d0101 GROUP by MODE" % (log.name, year, year+1)
            c = self.logbook.connection.cursor()
            c.execute(query)
            xy = c.fetchall()

            for i in range(len(xy)):
                mode = xy[i][0]
                if mode == "":
                    mode = "Unspecified"

                # Add to running total
                if mode in mode_count.keys():
                    mode_count[mode] += xy[i][1]
                else:
                    mode_count[mode] = xy[i][1]

        return mode_count

    def update(self):
        """ Update the information presented on the summary page. """

        self.items["LOG_COUNT"].set_label(str(self.logbook.log_count))
        self.items["QSO_COUNT"].set_label(str(self.logbook.record_count))
        try:
            t = datetime.fromtimestamp(getmtime(self.logbook.path)).strftime("%d %B %Y @ %H:%M")
            self.items["DATE_MODIFIED"].set_label(str(t))
        except (IOError, OSError) as e:
            logging.exception(e)
        return



#!/usr/bin/env python3

#    Copyright (C) 2017 Christian Thomas Jacobs.

#    This file is part of PyQSO.

#    PyQSO is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyQSO is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyQSO.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk
import unittest
try:
    import unittest.mock as mock
except ImportError:
    import mock
from pyqso.record_dialog import *


class TestRecordDialog(unittest.TestCase):

    """ The unit tests for the RecordDialog class. """

    def setUp(self):
        """ Set up the objects needed for the unit tests. """
        PyQSO = mock.MagicMock()
        self.record_dialog = RecordDialog(application=PyQSO(), log=None)
        self.record_dialog.frequency_unit = "MHz"

        # Set up the necessary sources.
        self.record_dialog.sources["FREQ"] = Gtk.Entry()

        self.record_dialog.sources["BAND"] = Gtk.ComboBoxText()
        for band in BANDS:
            self.record_dialog.sources["BAND"].append_text(band)

        self.record_dialog.sources["MODE"] = Gtk.ComboBoxText()
        for mode in sorted(MODES.keys()):
            self.record_dialog.sources["MODE"].append_text(mode)

        self.record_dialog.sources["SUBMODE"] = Gtk.ComboBoxText()
        self.record_dialog.sources["SUBMODE"].append_text("")
        self.record_dialog.sources["SUBMODE"].set_active(0)

        return

    def test_autocomplete_band(self):
        """ Given a frequency, check that the band field is automatically set to the correct value. """
        self.record_dialog.sources["FREQ"].set_text("145.525")
        self.record_dialog.autocomplete_band()
        band = self.record_dialog.sources["BAND"].get_active_text()
        assert(band == "2m")

        self.record_dialog.sources["FREQ"].set_text("9001")
        self.record_dialog.autocomplete_band()
        band = self.record_dialog.sources["BAND"].get_active_text()
        assert(band == "")  # Frequency does not lie in any of the specified bands.

    def test_convert_frequency(self):
        """ Check that a frequency can be successfully converted from one unit to another. """
        frequency = "7.140"  # In MHz
        converted = self.record_dialog.convert_frequency(frequency, from_unit="MHz", to_unit="AHz")  # Unknown to_unit. This should return the input unmodified (and give an error message).
        assert(converted == frequency)
        converted = self.record_dialog.convert_frequency(frequency, from_unit="MHz", to_unit="kHz")  # Convert from MHz to kHz.
        assert(float(converted) == 1e3*float(frequency))
        converted = self.record_dialog.convert_frequency(converted, from_unit="kHz", to_unit="MHz")  # Convert from kHz back to MHz. This should give the original frequency.
        assert(float(converted) == float(frequency))

        # Floating-point data type.
        frequency = 7.140  # In MHz
        converted = self.record_dialog.convert_frequency(frequency, from_unit="MHz", to_unit="kHz")
        assert(converted == frequency*1e3)

        # Floating-point data type.
        frequency = 7.140  # In MHz
        converted = self.record_dialog.convert_frequency(frequency, from_unit="MHz", to_unit="MHz")
        assert(converted == frequency)

        # Empty string.
        frequency = ""
        converted = self.record_dialog.convert_frequency(frequency, from_unit="MHz", to_unit="kHz")
        assert(converted == frequency)

        # Not a valid frequency.
        frequency = "HelloWorld"
        converted = self.record_dialog.convert_frequency(frequency, from_unit="MHz", to_unit="kHz")
        assert(converted == frequency)

    def test_hamlib_autofill(self):
        """ Check that FREQ, MODE and SUBMODE information can be retrieved from Hamlib's dummy rig (if the Hamlib module exists). """
        if(have_hamlib):
            rig_model = "RIG_MODEL_DUMMY"
            rig_pathname = "/dev/Rig"
            self.record_dialog.hamlib_autofill(rig_model, rig_pathname)
            assert(self.record_dialog.sources["FREQ"].get_text() == "145.000000")
            assert(self.record_dialog.sources["MODE"].get_active_text() == "FM")
            assert(self.record_dialog.sources["SUBMODE"].get_active_text() == "")
        else:
            pass

if(__name__ == '__main__'):
    unittest.main()



#!/usr/bin/env python3

#    Copyright (C) 2013-2017 Christian Thomas Jacobs.

#    This file is part of PyQSO.

#    PyQSO is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyQSO is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyQSO.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk
import logging
import sqlite3 as sqlite

from pyqso.adif import AVAILABLE_FIELD_NAMES_ORDERED


class Log(Gtk.ListStore):

    """ A single log inside of the whole logbook. A Log object can store multiple records. This is """

    def __init__(self, connection, name):
        """ Set up a new Log object.

        :arg connection: An sqlite database connection.
        :arg str name: The name of the log (i.e. the database table name).
        """

        # The ListStore constructor needs to know the data types of the columns.
        # The index is always an integer. We will assume the fields are strings.
        data_types = [int] + [str]*len(AVAILABLE_FIELD_NAMES_ORDERED)
        # Call the constructor of the super class (Gtk.ListStore).
        Gtk.ListStore.__init__(self, *data_types)

        self.connection = connection
        self.name = name

        return

    def populate(self):
        """ Remove everything in the Gtk.ListStore that is rendered already (via the TreeView), and start afresh. """

        logging.debug("Populating '%s'..." % self.name)
        self.add_missing_db_columns()
        self.clear()

        try:
            records = self.records

            for r in records:
                liststore_entry = [r["id"]]
                for field_name in AVAILABLE_FIELD_NAMES_ORDERED:
                    # Note: r may contain column names that are not in AVAILABLE_FIELD_NAMES_ORDERED,
                    # so we need to loop over and only select those that are, since the ListStore will
                    # expect a specific number of columns.
                    liststore_entry.append(r[field_name])
                self.append(liststore_entry)
            logging.debug("Finished populating '%s'." % self.name)

        except sqlite.Error as e:
            logging.error("Could not populate '%s' because of a database error." % self.name)
            logging.exception(e)

        return

    def add_missing_db_columns(self):
        """ Check whether each field name in AVAILABLE_FIELD_NAMES_ORDERED is in the database table. If not, add it
        (with all entries being set to an empty string initially).

        :raises sqlite.Error, IndexError: If the existing database column names could not be obtained, or missing column names could not be added.
        """
        logging.debug("Adding any missing database columns...")

        # Get all the column names in the current database table.
        column_names = []
        try:
            with self.connection:
                c = self.connection.cursor()
                c.execute("PRAGMA table_info(%s)" % self.name)
                result = c.fetchall()
            for t in result:
                column_names.append(t[1].upper())
        except (sqlite.Error, IndexError) as e:
            logging.exception(e)
            logging.error("Could not obtain the database column names.")
            return

        for field_name in AVAILABLE_FIELD_NAMES_ORDERED:
            if(not(field_name in column_names)):
                try:
                    with self.connection:
                        c.execute("ALTER TABLE %s ADD COLUMN %s TEXT DEFAULT \"\"" % (self.name, field_name.lower()))
                except sqlite.Error as e:
                    logging.exception(e)
                    logging.error("Could not add the missing database column '%s'." % field_name)
                    pass
        logging.debug("Finished adding any missing database columns.")
        return

    def add_record(self, fields_and_data):
        """ Add a record (or multiple records) to the log.

        :arg fields_and_data: A list of dictionaries (or possibly just a single dictionary), with each dictionary representing a single QSO, to be added to the log.
        """
        logging.debug("Adding record(s) to log...")

        # If a dictionary is given, assume that we only have one record to add.
        if isinstance(fields_and_data, dict):
            fields_and_data = [fields_and_data]

        with self.connection:
            c = self.connection.cursor()

            # Get all the column names in the current database table.
            c.execute("PRAGMA table_info(%s)" % self.name)
            column_names = c.fetchall()

            # Get the index/rowid of the last inserted record in the database.
            c.execute("SELECT max(id) FROM %s" % self.name)
            last_index = c.fetchone()[0]
            if last_index is None:
                # Assume no records are currently present.
                last_index = 0

        # A list of all the database entries, to be inserted in one go into the database.
        database_entries = []

        # Construct the SQL query.
        query = "INSERT INTO %s VALUES (NULL" % self.name
        for i in range(len(column_names)-1):  # -1 here because we don't want to count the database's 'id' column, since this is autoincremented.
            query = query + ",?"
        query = query + ")"

        # Gather all the records (making sure that the entries of each record are in the correct order).
        for r in range(len(fields_and_data)):
            # What if the database columns are not necessarily in the same order as (or even exist in) AVAILABLE_FIELD_NAMES_ORDERED?
            # PyQSO handles this here, but needs a separate list (called database_entry) to successfully perform the SQL query.
            database_entry = []
            for t in column_names:
                column_name = str(t[1])  # 't' here is a tuple
                if((column_name.upper() in AVAILABLE_FIELD_NAMES_ORDERED) and (column_name.upper() in list(fields_and_data[r].keys()))):
                    database_entry.append(fields_and_data[r][column_name.upper()])
                else:
                    if(column_name != "id"):  # Ignore the index/rowid field. This is a special case since it's not in AVAILABLE_FIELD_NAMES_ORDERED.
                        database_entry.append("")
            database_entries.append(database_entry)

        # Insert records in the database.
        with self.connection:
            c = self.connection.cursor()
            c.executemany(query, database_entries)

            # Get the indices/rowids of the newly-inserted records.
            query = "SELECT id FROM %s WHERE id > %s ORDER BY id ASC" % (self.name, last_index)
            c.execute(query)
            inserted = c.fetchall()

            # Check that the number of records we wanted to insert is the same as the number of records successfully inserted.
            assert(len(inserted) == len(database_entries))

            # Add the records to the ListStore as well.
            for r in range(len(fields_and_data)):
                liststore_entry = [inserted[r]["id"]]  # Add the record's index.
                field_names = AVAILABLE_FIELD_NAMES_ORDERED
                for i in range(0, len(field_names)):
                    if(field_names[i] in list(fields_and_data[r].keys())):
                        liststore_entry.append(fields_and_data[r][field_names[i]])
                    else:
                        liststore_entry.append("")
                self.append(liststore_entry)

        logging.debug("Successfully added the record(s) to the log.")
        return

    def delete_record(self, index, iter=None):
        """ Delete a specified record from the log. The corresponding record is also deleted from the Gtk.ListStore data structure.

        :arg int index: The index of the record in the SQL database.
        :arg iter: The iterator pointing to the record to be deleted in the Gtk.ListStore. If the default value of None is used, only the database entry is deleted and the corresponding Gtk.ListStore is left alone.
        :raises sqlite.Error, IndexError: If the record could not be deleted.
        """
        logging.debug("Deleting record from log...")

        # Delete the selected row in database.
        with self.connection:
            c = self.connection.cursor()
            query = "DELETE FROM %s" % self.name
            c.execute(query+" WHERE id=?", [index])

        # Delete the selected row in the Gtk.ListStore.
        if(iter is not None):
            self.remove(iter)

        logging.debug("Successfully deleted the record from the log.")
        return

    def edit_record(self, index, field_name, data, iter=None, column_index=None):
        """ Edit a specified record by replacing the current data in a specified field with the data provided.

        :arg int index: The index of the record in the SQL database.
        :arg str field_name: The name of the field whose data should be modified.
        :arg str data: The data that should replace the current data in the field.
        :arg iter: The iterator pointing to the record to be edited in the Gtk.ListStore. If the default value of None is used, only the database entry is edited and the corresponding Gtk.ListStore is left alone.
        :arg column_index: The index of the column in the Gtk.ListStore to be edited. If the default value of None is used, only the database entry is edited and the corresponding Gtk.ListStore is left alone.
        :raises sqlite.Error, IndexError: If the record could not be edited.
        """
        logging.debug("Editing field '%s' in record %d..." % (field_name, index))
        with self.connection:
            c = self.connection.cursor()
            query = "UPDATE %s SET %s" % (self.name, field_name)
            query = query + "=? WHERE id=?"
            c.execute(query, [data, index])  # First update the SQL database...
        if(iter is not None and column_index is not None):
            self.set(iter, column_index, data)  # ...and then the ListStore.
        logging.debug("Successfully edited field '%s' in record %d in the log." % (field_name, index))
        return

    def remove_duplicates(self):
        """ Remove any duplicate records from the log.

        :returns: The total number of duplicates, and the number of duplicates that were successfully removed. Hopefully these will be the same.
        :rtype: tuple
        """
        duplicates = self.get_duplicates()
        if(len(duplicates) == 0):
            return (0, 0)  # Nothing to do here.

        removed = 0  # Count the number of records that are removed. Hopefully this will be the same as len(duplicates).
        iter = self.get_iter_first()  # Start with the first row in the log.
        prev = iter  # Keep track of the previous iter (initially this will be the same as the first row in the log).
        while iter is not None:
            row_index = self.get_value(iter, 0)  # Get the index.
            if(row_index in duplicates):  # Is this a duplicate row? If so, delete it.
                self.delete_record(row_index, iter)
                removed += 1
                iter = prev  # Go back to the iter before the record that was just removed and continue from there.
                continue
            prev = iter
            iter = self.iter_next(iter)  # Move on to the next row, until iter_next returns None.

        return (len(duplicates), removed)

    def rename(self, new_name):
        """ Rename the log.

        :arg str new_name: The new name for the log.
        :returns: True if the renaming process is successful. Otherwise returns False.
        :rtype: bool
        """
        try:
            with self.connection:
                # First try to alter the table name in the database.
                c = self.connection.cursor()
                query = "ALTER TABLE %s RENAME TO %s" % (self.name, new_name)
                c.execute(query)
            # If the table name change was successful, then change the name attribute of the Log object too.
            self.name = new_name
            success = True
        except sqlite.Error as e:
            logging.exception(e)
            success = False
        return success

    def get_duplicates(self):
        """ Find the duplicates in the log, based on the CALL, QSO_DATE, and TIME_ON fields.

        :returns: A list of indices/ids corresponding to the duplicate records.
        :rtype: list
        """
        duplicates = []
        try:
            with self.connection:
                c = self.connection.cursor()
                c.execute(
                    """SELECT id FROM %s WHERE id NOT IN
   (
   SELECT MIN(id) FROM %s GROUP BY call, qso_date, time_on
   )""" % (self.name, self.name))
                result = c.fetchall()
            for index in result:
                duplicates.append(index[0])  # Get the integer from inside the tuple.
            duplicates.sort()  # These indices should monotonically increasing, but let's sort the list just in case.
        except (sqlite.Error, IndexError) as e:
            logging.exception(e)
        return duplicates

    def get_record_by_index(self, index):
        """ Return a record with a given index in the log.

        :arg int index: The index of the record in the SQL database.
        :returns: The desired record, represented by a dictionary of field-value pairs.
        :rtype: dict
        :raises sqlite.Error: If the record could not be retrieved from the database.
        """
        with self.connection:
            c = self.connection.cursor()
            query = "SELECT * FROM %s WHERE id=?" % self.name
            c.execute(query, [index])
            return c.fetchone()

    @property
    def records(self):
        """ Return a list of all the records in the log.

        :returns: A list of all the records in the log. Each record is represented by a dictionary.
        :rtype: dict
        :raises sqlite.Error: If the records could not be retrieved from the database.
        """
        with self.connection:
            c = self.connection.cursor()
            c.execute("SELECT * FROM %s" % self.name)
            return c.fetchall()

    @property
    def record_count(self):
        """ Return the total number of records in the log.

        :returns: The total number of records in the log.
        :rtype: int
        :raises sqlite.Error: If the record count could not be determined due to a database error.
        """
        with self.connection:
            c = self.connection.cursor()
            c.execute("SELECT Count(*) FROM %s" % self.name)
            return c.fetchone()[0]



#!/usr/bin/env python3

#    Copyright (C) 2013-2017 Christian Thomas Jacobs.

#    This file is part of PyQSO.

#    PyQSO is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyQSO is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyQSO.  If not, see <http://www.gnu.org/licenses/>.


class Toolbar:

    """ The toolbar underneath the menu bar. """

    def __init__(self, application):
        """ Set up the various buttons in the toolbar, and connect to their corresponding functions.

        :arg application: The PyQSO application containing the main Gtk window, etc.
        """

        self.application = application
        self.builder = self.application.builder

        self.buttons = {}

        # Create logbook
        self.buttons["NEW_LOGBOOK"] = self.builder.get_object("toolbar_new_logbook")
        self.buttons["NEW_LOGBOOK"].connect("clicked", self.application.logbook.new)

        # Open logbook
        self.buttons["OPEN_LOGBOOK"] = self.builder.get_object("toolbar_open_logbook")
        self.buttons["OPEN_LOGBOOK"].connect("clicked", self.application.logbook.open)

        # Close logbook
        self.buttons["CLOSE_LOGBOOK"] = self.builder.get_object("toolbar_close_logbook")
        self.buttons["CLOSE_LOGBOOK"].connect("clicked", self.application.logbook.close)

        # Add record
        self.buttons["ADD_RECORD"] = self.builder.get_object("toolbar_add_record")
        self.buttons["ADD_RECORD"].connect("clicked", self.application.logbook.add_record_callback)

        # Edit record
        self.buttons["EDIT_RECORD"] = self.builder.get_object("toolbar_edit_record")
        self.buttons["EDIT_RECORD"].connect("clicked", self.application.logbook.edit_record_callback)

        # Delete record
        self.buttons["DELETE_RECORD"] = self.builder.get_object("toolbar_delete_record")
        self.buttons["DELETE_RECORD"].connect("clicked", self.application.logbook.delete_record_callback)

        # Filter log
        self.filter_source = self.builder.get_object("filter_source")
        self.filter_source.connect_after("changed", self.application.logbook.filter_logs)

        # Set sensitivities.
        self.set_logbook_button_sensitive(True)
        self.set_record_buttons_sensitive(False)
        self.filter_source.set_sensitive(False)

        return

    def set_logbook_button_sensitive(self, sensitive):
        """ Enable/disable logbook-related toolbar items.

        :arg bool sensitive: If True, enable the 'new logbook' and 'open logbook' toolbar items. If False, disable them.
        """
        self.buttons["NEW_LOGBOOK"].set_sensitive(sensitive)
        self.buttons["OPEN_LOGBOOK"].set_sensitive(sensitive)
        self.buttons["CLOSE_LOGBOOK"].set_sensitive(not sensitive)
        return

    def set_record_buttons_sensitive(self, sensitive):
        """ Enable/disable record-related toolbar items.

        :arg bool sensitive: If True, enable all the record-related toolbar items. If False, disable them all.
        """
        for button_name in ["ADD_RECORD", "EDIT_RECORD", "DELETE_RECORD"]:
            self.buttons[button_name].set_sensitive(sensitive)
        return



#!/usr/bin/env python3

#    Copyright (C) 2013-2018 Christian Thomas Jacobs.

#    This file is part of PyQSO.

#    PyQSO is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyQSO is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyQSO.  If not, see <http://www.gnu.org/licenses/>.

from setuptools import setup

setup(name="PyQSO",
      version="1.1.0",
      description="A contact logging tool for amateur radio operators.",
      author="Christian Thomas Jacobs",
      author_email="christian@christianjacobs.uk",
      url="https://github.com/ctjacobs/pyqso",
      classifiers=[
          "Development Status :: 5 - Production/Stable",
          "Intended Audience :: End Users/Desktop",
          "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
          "Natural Language :: English",
          "Programming Language :: Python :: 3",
          "Topic :: Communications :: Ham Radio",
      ],
      packages=["pyqso"],
      package_dir={"pyqso": "pyqso"},
      package_data={"pyqso": ["res/pyqso.glade", "res/log_64x64.png"]},
      scripts=["bin/pyqso"],
      zip_safe=False
      )



#!/usr/bin/env python3

#    Copyright (C) 2013-2017 Christian Thomas Jacobs.

#    This file is part of PyQSO.

#    PyQSO is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyQSO is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyQSO.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk
import logging
try:
    import configparser
except ImportError:
    import ConfigParser as configparser
import os.path
import base64
try:
    import Hamlib
    have_hamlib = True
except ImportError:
    logging.warning("Could not import the Hamlib module!")
    have_hamlib = False
try:
    import geocoder
    have_geocoder = True
except ImportError:
    logging.warning("Could not import the geocoder module!")
    have_geocoder = False

from pyqso.adif import AVAILABLE_FIELD_NAMES_ORDERED, MODES
from pyqso.auxiliary_dialogs import error

PREFERENCES_FILE = os.path.expanduser("~/.config/pyqso/preferences.ini")


class PreferencesDialog:

    """ A dialog to specify the PyQSO preferences. """

    def __init__(self, application):
        """ Set up the various pages of the preferences dialog.

        :arg application: The PyQSO application containing the main Gtk window, etc.
        """

        logging.debug("Setting up the preferences dialog...")

        self.application = application
        self.builder = self.application.builder
        glade_file_path = os.path.join(os.path.realpath(os.path.dirname(__file__)), "res", "pyqso.glade")
        self.builder.add_objects_from_file(glade_file_path, ("preferences_dialog",))
        self.dialog = self.builder.get_object("preferences_dialog")

        self.general = GeneralPage(self.dialog, self.builder)
        self.view = ViewPage(self.dialog, self.builder)
        self.records = RecordsPage(self.dialog, self.builder)
        self.import_export = ImportExportPage(self.dialog, self.builder)
        self.hamlib = HamlibPage(self.dialog, self.builder)
        self.world_map = WorldMapPage(self.dialog, self.builder)

        self.dialog.show_all()

        logging.debug("Preferences dialog ready!")

        return

    def commit(self):
        """ Commit the user preferences to the configuration file. """

        logging.debug("Committing the user preferences to the configuration file...")

        config = configparser.ConfigParser()

        # General
        config.add_section("general")
        for key in list(self.general.data.keys()):
            config.set("general", key.lower(), str(self.general.data[key]))

        # View
        config.add_section("view")
        for key in list(self.view.data.keys()):
            config.set("view", key.lower(), str(self.view.data[key]))

        # Records
        config.add_section("records")
        for key in list(self.records.data.keys()):
            config.set("records", key.lower(), str(self.records.data[key]))

        # Import/Export
        config.add_section("import_export")
        for key in list(self.import_export.data.keys()):
            config.set("import_export", key.lower(), str(self.import_export.data[key]))

        # Hamlib
        config.add_section("hamlib")
        for key in list(self.hamlib.data.keys()):
            config.set("hamlib", key.lower(), str(self.hamlib.data[key]))

        # World Map
        config.add_section("world_map")
        for key in list(self.world_map.data.keys()):
            config.set("world_map", key.lower(), str(self.world_map.data[key]))

        # Write the preferences to file.
        with open(os.path.expanduser(PREFERENCES_FILE), 'w') as f:
            config.write(f)

        return


class GeneralPage:

    """ The section of the preferences dialog containing general preferences. """

    def __init__(self, parent, builder):
        """ Set up the General page of the Preferences dialog. """

        self.parent = parent
        self.builder = builder
        self.sources = {}

        # Remember that the have_config conditional in the PyQSO class may be out-of-date the next time the user opens up the preferences dialog
        # because a configuration file may have been created after launching the application. Let's check to see if one exists again...
        config = configparser.ConfigParser()
        have_config = (config.read(PREFERENCES_FILE) != [])

        # Show toolbox.
        self.sources["SHOW_TOOLBOX"] = self.builder.get_object("general_show_toolbox_checkbutton")
        (section, option) = ("general", "show_toolbox")
        if(have_config and config.has_option(section, option)):
            self.sources["SHOW_TOOLBOX"].set_active(config.getboolean(section, option))
        else:
            self.sources["SHOW_TOOLBOX"].set_active(False)

        # Show statistics.
        self.sources["SHOW_YEARLY_STATISTICS"] = self.builder.get_object("general_show_yearly_statistics_checkbutton")
        (section, option) = ("general", "show_yearly_statistics")
        if(have_config and config.has_option(section, option)):
            self.sources["SHOW_YEARLY_STATISTICS"].set_active(config.getboolean(section, option))
        else:
            self.sources["SHOW_YEARLY_STATISTICS"].set_active(False)

        # Default logbook.
        self.sources["DEFAULT_LOGBOOK"] = self.builder.get_object("general_default_logbook_checkbutton")
        (section, option) = ("general", "default_logbook")
        if(have_config and config.has_option(section, option)):
            self.sources["DEFAULT_LOGBOOK"].set_active(config.getboolean(section, option))
        else:
            self.sources["DEFAULT_LOGBOOK"].set_active(False)
        self.sources["DEFAULT_LOGBOOK"].connect("toggled", self.on_default_logbook_toggled)

        self.sources["DEFAULT_LOGBOOK_PATH"] = self.builder.get_object("general_default_logbook_entry")
        (section, option) = ("general", "default_logbook")
        # Disable the text entry box if the default logbook checkbox is not checked.
        if(have_config and config.has_option(section, option)):
            self.sources["DEFAULT_LOGBOOK_PATH"].set_sensitive(self.sources["DEFAULT_LOGBOOK"].get_active())
            self.builder.get_object("general_default_logbook_button").set_sensitive(self.sources["DEFAULT_LOGBOOK"].get_active())
        else:
            self.sources["DEFAULT_LOGBOOK_PATH"].set_sensitive(False)
            self.builder.get_object("general_default_logbook_button").set_sensitive(False)
        (section, option) = ("general", "default_logbook_path")
        if(have_config and config.has_option(section, option)):
            self.sources["DEFAULT_LOGBOOK_PATH"].set_text(config.get(section, option))

        self.builder.get_object("general_default_logbook_button").connect("clicked", self.on_default_logbook_clicked)

        # Keep 'Add Record' dialog open.
        self.sources["KEEP_OPEN"] = self.builder.get_object("general_keep_open_checkbutton")
        (section, option) = ("general", "keep_open")
        if(have_config and config.has_option(section, option)):
            self.sources["KEEP_OPEN"].set_active(config.getboolean(section, option))
        else:
            self.sources["KEEP_OPEN"].set_active(False)

        return

    @property
    def data(self):
        """ User preferences regarding General settings. """
        data = {}
        data["SHOW_TOOLBOX"] = self.sources["SHOW_TOOLBOX"].get_active()
        data["SHOW_YEARLY_STATISTICS"] = self.sources["SHOW_YEARLY_STATISTICS"].get_active()
        data["DEFAULT_LOGBOOK"] = self.sources["DEFAULT_LOGBOOK"].get_active()
        data["DEFAULT_LOGBOOK_PATH"] = os.path.expanduser(self.sources["DEFAULT_LOGBOOK_PATH"].get_text())
        data["KEEP_OPEN"] = self.sources["KEEP_OPEN"].get_active()
        return data

    def on_default_logbook_toggled(self, widget, data=None):
        if(widget.get_active()):
            self.sources["DEFAULT_LOGBOOK_PATH"].set_sensitive(True)
            self.builder.get_object("general_default_logbook_button").set_sensitive(True)
        else:
            self.sources["DEFAULT_LOGBOOK_PATH"].set_sensitive(False)
            self.builder.get_object("general_default_logbook_button").set_sensitive(False)
        return

    def on_default_logbook_clicked(self, widget):
        """ Let the user select the default logbook file via a file chooser dialog,
        and set the path in the adjacent entry box. """

        dialog = Gtk.FileChooserDialog("Select SQLite Database File",
                                       self.parent,
                                       Gtk.FileChooserAction.OPEN,
                                       (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                                        Gtk.STOCK_OPEN, Gtk.ResponseType.OK))
        response = dialog.run()
        if(response == Gtk.ResponseType.OK):
            path = dialog.get_filename()
            self.sources["DEFAULT_LOGBOOK_PATH"].set_text(path)
        dialog.destroy()
        return


class ViewPage:

    """ The section of the preferences dialog containing view-related preferences. """

    def __init__(self, parent, builder):
        """ Set up the View page of the Preferences dialog. """

        self.parent = parent
        self.builder = builder
        self.sources = {}

        config = configparser.ConfigParser()
        have_config = (config.read(PREFERENCES_FILE) != [])

        # Visible fields
        for field_name in AVAILABLE_FIELD_NAMES_ORDERED:
            self.sources[field_name] = self.builder.get_object("visible_fields_%s" % (field_name.lower()))
            if(have_config and config.has_option("view", field_name.lower())):
                self.sources[field_name].set_active(config.getboolean("view", field_name.lower()))
            else:
                self.sources[field_name].set_active(True)

        return

    @property
    def data(self):
        """ User preferences regarding View settings. """
        data = {}
        for field_name in AVAILABLE_FIELD_NAMES_ORDERED:
            data[field_name] = self.sources[field_name].get_active()
        return data


class RecordsPage:

    """ The section of the preferences dialog containing record-related preferences. """

    def __init__(self, parent, builder):
        """ Set up the Record page of the Preferences dialog. """

        self.parent = parent
        self.builder = builder
        self.sources = {}

        # Remember that the have_config conditional in the PyQSO class may be out-of-date the next time the user opens up the preferences dialog
        # because a configuration file may have been created after launching the application. Let's check to see if one exists again...
        config = configparser.ConfigParser()
        have_config = (config.read(PREFERENCES_FILE) != [])

        # Autocomplete
        self.sources["AUTOCOMPLETE_BAND"] = self.builder.get_object("records_autocomplete_band_checkbutton")
        (section, option) = ("records", "autocomplete_band")
        if(have_config and config.has_option(section, option)):
            self.sources["AUTOCOMPLETE_BAND"].set_active(config.getboolean(section, option))
        else:
            self.sources["AUTOCOMPLETE_BAND"].set_active(True)

        self.sources["USE_UTC"] = self.builder.get_object("records_autocomplete_utc_checkbutton")
        (section, option) = ("records", "use_utc")
        if(have_config and config.has_option(section, option)):
            self.sources["USE_UTC"].set_active(config.getboolean(section, option))
        else:
            self.sources["USE_UTC"].set_active(True)

        # Default values

        # Mode
        self.sources["DEFAULT_MODE"] = self.builder.get_object("default_values_mode_combo")
        for mode in sorted(MODES.keys()):
            self.sources["DEFAULT_MODE"].append_text(mode)
        (section, option) = ("records", "default_mode")
        if(have_config and config.has_option(section, option)):
            mode = config.get(section, option)
        else:
            mode = ""
        self.sources["DEFAULT_MODE"].set_active(sorted(MODES.keys()).index(mode))
        self.sources["DEFAULT_MODE"].connect("changed", self.on_mode_changed)

        # Submode
        self.sources["DEFAULT_SUBMODE"] = self.builder.get_object("default_values_submode_combo")
        for submode in MODES[mode]:
            self.sources["DEFAULT_SUBMODE"].append_text(submode)
        (section, option) = ("records", "default_submode")
        if(have_config and config.has_option(section, option)):
            submode = config.get(section, option)
        else:
            submode = ""
        self.sources["DEFAULT_SUBMODE"].set_active(MODES[mode].index(submode))

        # Power
        self.sources["DEFAULT_POWER"] = self.builder.get_object("default_values_tx_power_entry")
        (section, option) = ("records", "default_power")
        if(have_config and config.has_option(section, option)):
            self.sources["DEFAULT_POWER"].set_text(config.get(section, option))
        else:
            self.sources["DEFAULT_POWER"].set_text("")

        # Frequency unit
        self.sources["DEFAULT_FREQUENCY_UNIT"] = self.builder.get_object("default_values_frequency_unit_combo")
        units = ["Hz", "kHz", "MHz", "GHz"]
        for unit in units:
            self.sources["DEFAULT_FREQUENCY_UNIT"].append_text(unit)
        (section, option) = ("records", "default_frequency_unit")
        if(have_config and config.has_option(section, option)):
            self.sources["DEFAULT_FREQUENCY_UNIT"].set_active(units.index(config.get(section, option)))
        else:
            self.sources["DEFAULT_FREQUENCY_UNIT"].set_active(units.index("MHz"))

        # Callsign lookup
        self.sources["CALLSIGN_DATABASE"] = self.builder.get_object("callsign_lookup_database_combo")
        callsign_database = ["", "qrz.com", "hamqth.com"]
        for database in callsign_database:
            self.sources["CALLSIGN_DATABASE"].append_text(database)
        (section, option) = ("records", "callsign_database")
        if(have_config and config.has_option(section, option)):
            self.sources["CALLSIGN_DATABASE"].set_active(callsign_database.index(config.get(section, option)))
        else:
            self.sources["CALLSIGN_DATABASE"].set_active(callsign_database.index(""))

        # Login details
        self.sources["CALLSIGN_DATABASE_USERNAME"] = self.builder.get_object("callsign_lookup_login_details_username_entry")
        (section, option) = ("records", "callsign_database_username")
        if(have_config and config.has_option(section, option)):
            self.sources["CALLSIGN_DATABASE_USERNAME"].set_text(config.get(section, option))

        self.sources["CALLSIGN_DATABASE_PASSWORD"] = self.builder.get_object("callsign_lookup_login_details_password_entry")
        (section, option) = ("records", "callsign_database_password")
        if(have_config and config.has_option(section, option)):
            password = base64.b64decode(config.get(section, option)).decode("utf-8")
            self.sources["CALLSIGN_DATABASE_PASSWORD"].set_text(password)

        self.sources["IGNORE_PREFIX_SUFFIX"] = self.builder.get_object("callsign_lookup_ignore_prefix_suffix_checkbutton")
        (section, option) = ("records", "ignore_prefix_suffix")
        if(have_config and config.has_option(section, option)):
            self.sources["IGNORE_PREFIX_SUFFIX"].set_active(config.getboolean(section, option))
        else:
            self.sources["IGNORE_PREFIX_SUFFIX"].set_active(True)

        return

    @property
    def data(self):
        """ User preferences regarding Records settings. """
        data = {}
        data["AUTOCOMPLETE_BAND"] = self.sources["AUTOCOMPLETE_BAND"].get_active()
        data["USE_UTC"] = self.sources["USE_UTC"].get_active()

        data["DEFAULT_MODE"] = self.sources["DEFAULT_MODE"].get_active_text()
        data["DEFAULT_SUBMODE"] = self.sources["DEFAULT_SUBMODE"].get_active_text()
        data["DEFAULT_POWER"] = self.sources["DEFAULT_POWER"].get_text()
        data["DEFAULT_FREQUENCY_UNIT"] = self.sources["DEFAULT_FREQUENCY_UNIT"].get_active_text()

        data["CALLSIGN_DATABASE"] = self.sources["CALLSIGN_DATABASE"].get_active_text()
        data["CALLSIGN_DATABASE_USERNAME"] = self.sources["CALLSIGN_DATABASE_USERNAME"].get_text()
        data["CALLSIGN_DATABASE_PASSWORD"] = base64.b64encode(self.sources["CALLSIGN_DATABASE_PASSWORD"].get_text().encode("utf-8")).decode("utf-8")  # Need to convert from bytes to str here.
        data["IGNORE_PREFIX_SUFFIX"] = self.sources["IGNORE_PREFIX_SUFFIX"].get_active()
        return data

    def on_mode_changed(self, combo):
        """ If the MODE field has changed its value, then fill the SUBMODE field with all the available SUBMODE options for that new MODE. """
        self.sources["DEFAULT_SUBMODE"].get_model().clear()
        mode = combo.get_active_text()
        for submode in MODES[mode]:
            self.sources["DEFAULT_SUBMODE"].append_text(submode)
        self.sources["DEFAULT_SUBMODE"].set_active(MODES[mode].index(""))
        return


class ImportExportPage:

    """ The section of the preferences dialog containing import/export-related preferences. """

    def __init__(self, parent, builder):
        """ Set up the Import/Export page of the Preferences dialog. """

        self.parent = parent
        self.builder = builder
        self.sources = {}

        # Remember that the have_config conditional in the PyQSO class may be out-of-date the next time the user opens up the preferences dialog
        # because a configuration file may have been created after launching the application. Let's check to see if one exists again...
        config = configparser.ConfigParser()
        have_config = (config.read(PREFERENCES_FILE) != [])

        # Import
        self.sources["MERGE_COMMENT"] = self.builder.get_object("adif_import_merge_comment_checkbutton")
        (section, option) = ("import_export", "merge_comment")
        if(have_config and config.has_option(section, option)):
            self.sources["MERGE_COMMENT"].set_active(config.getboolean(section, option))
        else:
            self.sources["MERGE_COMMENT"].set_active(False)

        return

    @property
    def data(self):
        """ User preferences regarding Import/Export settings. """
        data = {}
        data["MERGE_COMMENT"] = self.sources["MERGE_COMMENT"].get_active()
        return data


class HamlibPage:

    """ The section of the preferences dialog containing Hamlib-related preferences. """

    def __init__(self, parent, builder):
        """ Set up the Hamlib page of the Preferences dialog. """

        self.parent = parent
        self.builder = builder
        self.sources = {}

        config = configparser.ConfigParser()
        have_config = (config.read(PREFERENCES_FILE) != [])

        self.sources["AUTOFILL"] = self.builder.get_object("hamlib_support_checkbutton")
        (section, option) = ("hamlib", "autofill")
        if(have_config and config.has_option(section, option)):
            self.sources["AUTOFILL"].set_active(config.getboolean(section, option))
        else:
            self.sources["AUTOFILL"].set_active(False)

        # Get the list of rig models
        models = ["RIG_MODEL_NONE"]
        if(have_hamlib):
            try:
                for item in dir(Hamlib):
                    if(item.startswith("RIG_MODEL_")):
                        models.append(item)
            except:
                logging.error("Could not obtain rig models list via Hamlib!")
        else:
            logging.debug("Hamlib module not present. Could not obtain a list of rig models.")

        self.sources["RIG_MODEL"] = self.builder.get_object("hamlib_support_model_combo")
        for model in models:
            self.sources["RIG_MODEL"].append_text(model)
        (section, option) = ("hamlib", "rig_model")
        if(have_config and config.has_option("hamlib", "rig_model")):
            self.sources["RIG_MODEL"].set_active(models.index(config.get("hamlib", "rig_model")))
        else:
            self.sources["RIG_MODEL"].set_active(models.index("RIG_MODEL_NONE"))  # Set to RIG_MODEL_NONE as the default option.

        # Path to rig
        self.sources["RIG_PATHNAME"] = self.builder.get_object("hamlib_support_path_entry")
        (section, option) = ("hamlib", "rig_pathname")
        if(have_config and config.has_option(section, option)):
            self.sources["RIG_PATHNAME"].set_text(config.get(section, option))

        return

    @property
    def data(self):
        """ User preferences regarding Hamlib settings. """
        data = {}
        data["AUTOFILL"] = self.sources["AUTOFILL"].get_active()
        data["RIG_PATHNAME"] = self.sources["RIG_PATHNAME"].get_text()
        data["RIG_MODEL"] = self.sources["RIG_MODEL"].get_active_text()
        return data


class WorldMapPage:

    """ The section of the preferences dialog containing World Map preferences. """

    def __init__(self, parent, builder):
        """ Set up the World Map page of the Preferences dialog. """

        self.parent = parent
        self.builder = builder
        self.sources = {}

        # Remember that the have_config conditional in the PyQSO class may be out-of-date the next time the user opens up the preferences dialog
        # because a configuration file may have been created after launching the application. Let's check to see if one exists again...
        config = configparser.ConfigParser()
        have_config = (config.read(PREFERENCES_FILE) != [])

        # Option to pinpoint QTH on grey line map.
        self.sources["SHOW_QTH"] = self.builder.get_object("world_map_show_qth_checkbutton")
        (section, option) = ("world_map", "show_qth")
        if(have_config and config.has_option(section, option)):
            self.sources["SHOW_QTH"].set_active(config.getboolean(section, option))
        else:
            self.sources["SHOW_QTH"].set_active(False)

        self.sources["QTH_NAME"] = self.builder.get_object("world_map_qth_name_entry")
        button = self.builder.get_object("world_map_qth_lookup")
        button.connect("clicked", self.lookup_callback)  # Uses geocoding to find the latitude-longitude coordinates.

        self.sources["QTH_LATITUDE"] = self.builder.get_object("world_map_qth_coordinates_latitude_entry")
        self.sources["QTH_LONGITUDE"] = self.builder.get_object("world_map_qth_coordinates_longitude_entry")

        (section, option) = ("world_map", "show_qth")
        # Disable the text entry boxes if the SHOW_QTH checkbox is not checked.
        if(have_config and config.has_option(section, option)):
            self.sources["QTH_NAME"].set_sensitive(self.sources["SHOW_QTH"].get_active())
            self.sources["QTH_LATITUDE"].set_sensitive(self.sources["SHOW_QTH"].get_active())
            self.sources["QTH_LONGITUDE"].set_sensitive(self.sources["SHOW_QTH"].get_active())
            button.set_sensitive(self.sources["SHOW_QTH"].get_active())
        else:
            self.sources["QTH_NAME"].set_sensitive(False)
            self.sources["QTH_LATITUDE"].set_sensitive(False)
            self.sources["QTH_LONGITUDE"].set_sensitive(False)
            button.set_sensitive(False)
        (section, option) = ("world_map", "qth_name")
        if(have_config and config.has_option(section, option)):
            self.sources["QTH_NAME"].set_text(config.get(section, option))
        (section, option) = ("world_map", "qth_latitude")
        if(have_config and config.has_option(section, option)):
            self.sources["QTH_LATITUDE"].set_text(config.get(section, option))
        (section, option) = ("world_map", "qth_longitude")
        if(have_config and config.has_option(section, option)):
            self.sources["QTH_LONGITUDE"].set_text(config.get(section, option))
        self.sources["SHOW_QTH"].connect("toggled", self.on_show_qth_toggled)

        # Option to show Maidenhead grid squares.
        self.sources["SHOW_GRID_SQUARES"] = self.builder.get_object("world_map_show_grid_squares_checkbutton")
        (section, option) = ("world_map", "show_grid_squares")
        if(have_config and config.has_option(section, option)):
            self.sources["SHOW_GRID_SQUARES"].set_active(config.getboolean(section, option))
        else:
            self.sources["SHOW_GRID_SQUARES"].set_active(False)

        # Option to shade in worked Maidenhead grid squares.
        self.sources["SHADE_WORKED_GRID_SQUARES"] = self.builder.get_object("world_map_shade_worked_grid_squares_checkbutton")
        (section, option) = ("world_map", "shade_worked_grid_squares")
        if(have_config and config.has_option(section, option)):
            self.sources["SHADE_WORKED_GRID_SQUARES"].set_active(config.getboolean(section, option))
        else:
            self.sources["SHADE_WORKED_GRID_SQUARES"].set_active(False)

        return

    @property
    def data(self):
        """ User preferences regarding World Map settings. """
        data = {}
        data["SHOW_QTH"] = self.sources["SHOW_QTH"].get_active()
        data["QTH_NAME"] = self.sources["QTH_NAME"].get_text()
        data["QTH_LATITUDE"] = self.sources["QTH_LATITUDE"].get_text()
        data["QTH_LONGITUDE"] = self.sources["QTH_LONGITUDE"].get_text()
        data["SHOW_GRID_SQUARES"] = self.sources["SHOW_GRID_SQUARES"].get_active()
        data["SHADE_WORKED_GRID_SQUARES"] = self.sources["SHADE_WORKED_GRID_SQUARES"].get_active()
        return data

    def on_show_qth_toggled(self, widget, data=None):
        if(widget.get_active()):
            self.sources["QTH_NAME"].set_sensitive(True)
            self.sources["QTH_LATITUDE"].set_sensitive(True)
            self.sources["QTH_LONGITUDE"].set_sensitive(True)
            self.builder.get_object("world_map_qth_lookup").set_sensitive(True)
        else:
            self.sources["QTH_NAME"].set_sensitive(False)
            self.sources["QTH_LATITUDE"].set_sensitive(False)
            self.sources["QTH_LONGITUDE"].set_sensitive(False)
            self.builder.get_object("world_map_qth_lookup").set_sensitive(False)
        return

    def lookup_callback(self, widget=None):
        """ Perform geocoding of the QTH location to obtain latitude-longitude coordinates. """
        if(not have_geocoder):
            error(parent=self.parent, message="Geocoder module could not be imported. Geocoding aborted.")
            return
        logging.debug("Geocoding QTH location...")
        name = self.sources["QTH_NAME"].get_text()
        try:
            g = geocoder.google(name)
            latitude, longitude = g.latlng
            self.sources["QTH_LATITUDE"].set_text(str(latitude))
            self.sources["QTH_LONGITUDE"].set_text(str(longitude))
            logging.debug("QTH coordinates found: (%s, %s)", str(latitude), str(longitude))
        except ValueError as e:
            error(parent=self.parent, message="Unable to lookup QTH coordinates. Is the QTH name correct?")
            logging.exception(e)
        except Exception as e:
            error(parent=self.parent, message="Unable to lookup QTH coordinates. Check connection to the internets? Lookup limit reached?")
            logging.exception(e)
        return
